import sys
import unittest
import itertools
from fnmatch import fnmatchcase

import numpy as np
from scipy.sparse import coo_matrix


try:
    import jax
    import jax.numpy as jnp
except ImportError:
    jax = None

from openmdao.api import Problem, Group, IndepVarComp, ImplicitComponent, ExplicitComponent, \
    NonlinearBlockGS, ScipyOptimizeDriver, NewtonSolver, DirectSolver, ImplicitFuncComp
from openmdao.core.problem import _clear_problem_names
import openmdao.func_api as omf
from openmdao.utils.assert_utils import assert_near_equal, assert_warning
from openmdao.utils.array_utils import evenly_distrib_idxs
from openmdao.utils.testing_utils import require_pyoptsparse, use_tempdirs
from openmdao.utils.mpi import MPI
from openmdao.utils.coloring import compute_total_coloring


try:
    from openmdao.parallel_api import PETScVector
except ImportError:
    PETScVector = None

try:
    from parameterized import parameterized
except ImportError:
    from openmdao.utils.assert_utils import SkipParameterized as parameterized

from openmdao.utils.general_utils import set_pyoptsparse_opt, om_dump


# check that pyoptsparse is installed. if it is, try to use SLSQP.
OPT, OPTIMIZER = set_pyoptsparse_opt('SLSQP')
if OPTIMIZER:
    from openmdao.drivers.pyoptsparse_driver import pyOptSparseDriver
else:
    pyOptSparseDriver = None


def setup_vars(comp, ofs, wrts, sparse_partials=False, bad_sparsity=False):
    matrix = comp.sparsity
    isplit = comp.isplit
    osplit = comp.osplit

    slices = {}
    isizes, _ = evenly_distrib_idxs(isplit, matrix.shape[1])
    start = end = 0
    for i, sz in enumerate(isizes):
        end += sz
        wrt = f"x{i}"
        comp.add_input(wrt, np.zeros(sz))
        slices[wrt] = slice(start, end)
        start = end

    osizes, _ = evenly_distrib_idxs(osplit, matrix.shape[0])
    start = end = 0
    for i, sz in enumerate(osizes):
        end += sz
        of = f"y{i}"
        comp.add_output(of, np.zeros(sz))
        slices[of] = slice(start, end)
        start = end

    if sparse_partials:
        nbad = 0
        for i in range(len(osizes)):
            of = f"y{i}"
            for j in range(len(isizes)):
                wrt = f"x{j}"
                subjac = comp.sparsity[slices[of], slices[wrt]]
                if np.any(subjac):
                    rows, cols = np.nonzero(subjac)
                    if bad_sparsity and rows.size > 1 and nbad < 5:
                        rows[rows.size // 2] = -1  # remove one row/col pair
                        mask = rows != -1
                        rows = rows[mask]
                        cols = cols[mask]
                        nbad += 1
                    comp.declare_partials(of=of, wrt=wrt, rows=rows, cols=cols, method=comp.method)
                else:
                    comp.declare_partials(of=of, wrt=wrt, method=comp.method, dependent=False)
    else:
        comp.declare_partials(of=ofs, wrt=wrts, method=comp.method)


def setup_sparsity(mask):
    return np.random.random(mask.shape) * mask


def setup_indeps(isplit, ninputs, indeps_name, comp_name):
    isizes, _ = evenly_distrib_idxs(isplit, ninputs)
    indeps = IndepVarComp()
    conns = []
    for i, sz in enumerate(isizes):
        indep_var = 'x%d' % i
        indeps.add_output(indep_var, np.random.random(sz))
        conns.append((indeps_name + '.' + indep_var, comp_name + '.' + indep_var))

    return indeps, conns


class CounterGroup(Group):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._nruns = 0

    def _solve_nonlinear(self, *args, **kwargs):
        super()._solve_nonlinear(*args, **kwargs)
        self._nruns += 1


class SparseCompImplicit(ImplicitComponent):

    def __init__(self, sparsity, method='fd', isplit=1, osplit=1, **kwargs):
        super().__init__(**kwargs)
        self.sparsity = sparsity
        self.isplit = isplit
        self.osplit = osplit
        self.method = method
        self._nruns = 0

    def setup(self):
        setup_vars(self, ofs='*', wrts='*')

    # this is defined for easier testing of coloring of approx partials
    def apply_nonlinear(self, inputs, outputs, residuals):
        prod = self.sparsity.dot(inputs.asarray()) - outputs.asarray()
        start = end = 0
        for i in range(self.osplit):
            outname = 'y%d' % i
            end += outputs[outname].size
            residuals[outname] = prod[start:end]
            start = end
        self._nruns += 1

    # this is defined so we can more easily test coloring of approx totals in a Group above this comp
    def solve_nonlinear(self, inputs, outputs):
        prod = self.sparsity.dot(inputs.asarray())
        start = end = 0
        for i in range(self.osplit):
            outname = 'y%d' % i
            end += outputs[outname].size
            outputs[outname] = prod[start:end]
            start = end
        self._nruns += 1


# needed to wrap the sparsity array in this class in order to make it hashable so jax would be happy
class SpObj(object):
    def __init__(self, sparsity):
        self.mat = sparsity
        self._nruns = 0


class SparseFuncCompImplicit(ImplicitFuncComp):

    def __init__(self, sparsity, method='cs', isplit=1, osplit=1, **kwargs):
        self.sparsity = sparsity
        self.isplit = isplit
        self.osplit = osplit
        self.method = method
        self._nruns = 0

        isizes, _ = evenly_distrib_idxs(self.isplit, self.sparsity.shape[1])
        wrts = [(f"x{i}", sz) for i, sz in enumerate(isizes)]

        osizes, _ = evenly_distrib_idxs(self.osplit, self.sparsity.shape[0])
        ofs = [(f"y{i}", sz) for i, sz in enumerate(osizes)]

        # define apply_nonlinear function
        iparams = ','.join([n for n,_ in wrts])
        oparams = ','.join([n for n,_ in ofs])
        params = ','.join((iparams, oparams))
        flines = ['def func(' + params + ',sparsity):']
        if len(wrts) > 1:
            flines.append(f'    iarr = np.concatenate(({iparams}))')
        else:
            flines.append(f'    iarr = {iparams[0]}0')
        if len(ofs) > 1:
            flines.append(f'    oarr = np.concatenate(({oparams}))')
        else:
            flines.append(f'    oarr = {oparams[0]}0')
        flines.append('    prod = sparsity.mat.dot(iarr) - oarr')
        flines.append('    sparsity._nruns += 1')
        start = end = 0
        for name, sz in ofs:
            end += sz
            flines.append(f"    r{name} = prod[{start}:{end}]")
            start = end
        flines.append('    return ' + ','.join([f"r{n}" for n,_ in ofs]))
        fbody = '\n'.join(flines)

        if sys.version_info >= (3,13):
            # https://docs.python.org/3/whatsnew/3.13.html#pep667-porting-notes-py
            # Code execution functions that implicitly target locals() (such as exec and eval)
            # must be passed an explicit namespace to access their results in an optimized scope.
            tmp_locals = dict()
            exec(fbody, locals=tmp_locals)  # nosec trusted input
            f = omf.wrap(tmp_locals['func'])
        else:
            exec(fbody)  # nosec trusted input
            f = omf.wrap(locals()['func'])

        for name, sz in wrts:
            f.add_input(name, shape=sz)
        for name, sz in ofs:
            f.add_output(name, shape=sz, resid=f"r{name}")
        f.declare_option('sparsity', default=SpObj(sparsity))
        f.declare_partials(of='*', wrt='*', method=self.method)

        super().__init__(f, **kwargs)

    # this is defined for easier testing of coloring of approx partials
    def apply_nonlinear(self, inputs, outputs, residuals):
        super().apply_nonlinear(inputs, outputs, residuals)
        self._nruns += 1

    # this is defined so we can more easily test coloring of approx totals in a Group above this comp
    def solve_nonlinear(self, inputs, outputs):
        prod = self.sparsity.dot(inputs.asarray())
        start = end = 0
        for i in range(self.osplit):
            outname = 'y%d' % i
            end += outputs[outname].size
            outputs[outname] = prod[start:end]
            start = end
        self._nruns += 1


class SparseCompExplicit(ExplicitComponent):

    def __init__(self, sparsity, method='fd', isplit=1, osplit=1, sparse_partials=False, bad_sparsity=False, **kwargs):
        super().__init__(**kwargs)
        self.sparsity = sparsity
        self.isplit = isplit
        self.osplit = osplit
        self.method = method
        self.sparse_partials = sparse_partials
        self.bad_sparsity = bad_sparsity
        self._nruns = 0

    def setup(self):
        setup_vars(self, ofs='*', wrts='*', sparse_partials=self.sparse_partials, bad_sparsity=self.bad_sparsity)

    def compute(self, inputs, outputs):
        prod = self.sparsity.dot(inputs.asarray())
        start = end = 0
        for i in range(self.osplit):
            outname = 'y%d' % i
            end += outputs[outname].size
            outputs[outname] = prod[start:end]
            start = end
        self._nruns += 1


# relative tolerances for jacobian checks
_TOLS = {
    'fd': 1e-6,
    'cs': 1e-12,
    'jax': 1e-7
}

def _check_partial_matrix(system, jac, expected, method):
    blocks = []
    for abs_of, ofmeta in system._var_allprocs_abs2meta['output'].items():
        cblocks = []
        for abs_wrt, wrtmeta in system._var_allprocs_abs2meta['input'].items():
            key = (abs_of, abs_wrt)
            if key in jac:
                meta = jac[key]
                if meta['rows'] is not None:
                    cblocks.append(coo_matrix((meta['val'], (meta['rows'], meta['cols'])), shape=meta['shape']).toarray())
                elif meta['dependent']:
                    cblocks.append(meta['val'])
                else:
                    cblocks.append(np.zeros(meta['shape']))
            else: # sparsity was all zeros so we declared this subjac as not dependent
                relof = abs_of.rsplit('.', 1)[-1]
                relwrt = abs_wrt.rsplit('.', 1)[-1]
                for decl_key, meta in system._declared_partials_patterns.items():
                    if not meta['dependent']:
                        ofpats, wrtpats = decl_key
                        done = False
                        for p in ofpats:
                            if fnmatchcase(relof, p):
                                for wp in wrtpats:
                                    if fnmatchcase(relwrt, wp):
                                        cblocks.append(np.zeros((ofmeta['size'], wrtmeta['size'])))
                                        done = True
                                        break
                                if done:
                                    break
                if (relof, relwrt) in system._declared_partials_patterns and not system._declared_partials_patterns[(relof, relwrt)].get('dependent'):
                    cblocks.append(np.zeros((ofmeta['size'], wrtmeta['size'])))
        if cblocks:
            blocks.append(np.hstack(cblocks))
    fullJ = np.vstack(blocks)
    assert_near_equal(fullJ, expected, tolerance=_TOLS[method], tol_type='rel')


def _check_total_matrix(system, jac, expected, method):
    ofs = {}
    for key, subjac in jac.items():
        of, wrt = key
        if of not in ofs:
            ofs[of] = [subjac]
        else:
            ofs[of].append(subjac)

    for of, sjacs in ofs.items():
        ofs[of] = np.hstack(sjacs)

    fullJ = np.vstack(list(ofs.values()))
    assert_near_equal(fullJ, expected, tolerance=_TOLS[method], tol_type='rel')


def _check_semitotal_matrix(system, jac, expected, method):
    blocks = []
    for of in system._var_allprocs_abs2meta['output']:
        cblocks = []
        for wrt in itertools.chain(system._var_allprocs_abs2meta['output'], system._var_allprocs_abs2meta['input']):
            key = (of, wrt)
            if key in jac:
                rows = jac[key]['rows']
                if rows is not None:
                    cols = jac[key]['cols']
                    val = coo_matrix((jac[key]['val'], (rows, cols)), shape=jac[key]['shape']).toarray()
                else:
                    val = jac[key]['val']
                cblocks.append(val)
        if cblocks:
            blocks.append(np.hstack(cblocks))
    fullJ = np.vstack(blocks)
    assert_near_equal(fullJ, expected, tolerance=_TOLS[method], tol_type='rel')


_BIGMASK = np.array(
    [[1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1],
     [0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 1, 0, 1],
     [0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 1],
     [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0],
     [0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0],
     [0, 1, 0, 1, 1, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1],
     [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1],
     [0, 1, 0, 1, 0, 1, 1, 0, 1, 0, 1, 1, 0, 0, 1, 0, 0, 1, 0],
     [0, 1, 0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0],
     [1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0],
     [0, 1, 1, 0, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1]]
)

@use_tempdirs
class TestColoringExplicit(unittest.TestCase):
    def setUp(self):
        np.random.seed(11)

    def test_partials_explicit(self):

        for method, isplit, osplit, sparse_partials in itertools.product(['fd', 'cs'],
                                                                         [1,2,7,19],
                                                                         [1,2,5,11],
                                                                         [True, False]):

            with self.subTest(msg=f'{method=} {isplit=} {osplit=} {sparse_partials=}'):
                prob = Problem(name=f'test_partials_explicit_{method}_'
                               f'{isplit}_{osplit}_{sparse_partials}')
                model = prob.model

                sparsity = setup_sparsity(_BIGMASK)
                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'comp')
                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                                    isplit=isplit, osplit=osplit,
                                                                    sparse_partials=sparse_partials))
                comp.declare_coloring('x*', method=method)

                for conn in conns:
                    model.connect(*conn)

                prob.setup(mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                comp.run_linearize()
                prob.run_model()
                start_nruns = comp._nruns
                comp.run_linearize()
                self.assertEqual(comp._nruns - start_nruns, 10)
                jac = comp._jacobian._subjacs_info
                _check_partial_matrix(comp, jac, sparsity, method)

    def test_partials_explicit_static(self):
        for method, isplit, osplit in itertools.product(['fd', 'cs'],
                                                        [1,2,7,19],
                                                        [1,2,5,11]):

            with self.subTest(msg=f'{method=} {isplit=} {osplit=}'):
                probname = f'test_partials_explicit_static_{method}_{isplit}_{osplit}'
                prob = Problem(name=probname)
                model = prob.model

                sparsity = setup_sparsity(_BIGMASK)
                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'comp')
                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                                    isplit=isplit, osplit=osplit))
                for conn in conns:
                    model.connect(*conn)

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()
                coloring = comp._compute_coloring(wrt_patterns='x*', method=method)[0]
                comp._save_coloring(coloring)

                # This is a hack to get around the restriction of duplicate
                # problem names in the same execution of the same script.
                _clear_problem_names()

                # now make a second problem to use the coloring
                prob = Problem(name=probname)
                model = prob.model
                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                                    isplit=isplit, osplit=osplit))
                for conn in conns:
                    model.connect(*conn)

                comp.declare_coloring(wrt='x*', method=method)
                comp.use_fixed_coloring()
                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                start_nruns = comp._nruns
                comp._linearize()
                self.assertEqual(comp._nruns - start_nruns, 10)
                jac = comp._jacobian._subjacs_info
                _check_partial_matrix(comp, jac, sparsity, method)

    def test_partials_explicit_reuse(self):
        for num_insts in [1, 2, 5]:
            with self.subTest(msg=f'{num_insts=}'):
                method = 'cs'
                osplit = 5
                isplit = 7
                prob = Problem(name=f'test_partials_explicit_reuse_{num_insts}')
                model = prob.model

                sparsity = setup_sparsity(_BIGMASK)
                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'comp')
                model.add_subsystem('indeps', indeps)

                comps = []
                for i in range(num_insts):
                    cname = 'comp%d' % i
                    comp = model.add_subsystem(cname, SparseCompExplicit(sparsity, method,
                                                                        isplit=isplit, osplit=osplit))
                    comp.declare_coloring('x*', method=method, per_instance=False)
                    comps.append(comp)

                    _, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', cname)

                    for conn in conns:
                        model.connect(*conn)

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                prob.model.run_linearize()
                prob.run_model()
                start_nruns = [c._nruns for c in comps]
                for i, comp in enumerate(comps):
                    comp.run_linearize()
                    self.assertEqual(comp._nruns - start_nruns[i], 10)
                    jac = comp._jacobian._subjacs_info
                    _check_partial_matrix(comp, jac, sparsity, method)

                orig = comps[0]._coloring_info.coloring
                for comp in comps:
                    self.assertTrue(orig is comp._coloring_info.coloring,
                                    "Instance '{}' is using a different coloring".format(comp.pathname))


@use_tempdirs
class TestColoringImplicit(unittest.TestCase):
    def setUp(self):
        np.random.seed(11)

    def test_partials_implicit(self):
        for method, isplit, osplit in itertools.product(['fd', 'cs'],
                                                        [1,2,7,19],
                                                        [1,2,5,11]):
            with self.subTest(msg=f'{method=}_{isplit=}_{osplit=}'):
                prob = Problem(name=f'test_partials_implicit_{method}_{isplit}_{osplit}')
                model = prob.model

                sparsity = setup_sparsity(_BIGMASK)
                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'comp')
                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompImplicit(sparsity, method,
                                                                    isplit=isplit, osplit=osplit))
                comp.declare_coloring('x*', method=method)

                for conn in conns:
                    model.connect(*conn)

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                comp.run_linearize()
                prob.run_model()
                start_nruns = comp._nruns
                comp.run_linearize()
                self.assertEqual(comp._nruns - start_nruns, 10 + sparsity.shape[0])
                jac = comp._jacobian._subjacs_info
                _check_partial_matrix(comp, jac, sparsity, method)

    def test_simple_partials_implicit_static(self):
        for method, isplit, osplit in itertools.product(['fd', 'cs'],
                                                        [1,2,7,19],
                                                        [1,2,5,11]):
            with self.subTest(msg=f'{method=}_{isplit=}_{osplit=}'):
                prob = Problem(name=f'test_simple_partials_implicit_static_{method}_{isplit}_{osplit}')
                model = prob.model

                sparsity = setup_sparsity(_BIGMASK)
                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompImplicit(sparsity, method,
                                                                    isplit=isplit, osplit=osplit))
                for conn in conns:
                    model.connect(*conn)

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()
                coloring = comp._compute_coloring(wrt_patterns='x*', method=method)[0]
                comp._save_coloring(coloring)

                # now create a new problem and set the static coloring
                _clear_problem_names()  # We don't care about a duplicate problem name in this test.
                prob = Problem(name=f'test_simple_partials_implicit_static_{method}_{isplit}_{osplit}')
                model = prob.model

                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompImplicit(sparsity, method,
                                                                    isplit=isplit, osplit=osplit))
                for conn in conns:
                    model.connect(*conn)

                comp.declare_coloring(wrt='x*', method=method)
                comp.use_fixed_coloring()
                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                start_nruns = comp._nruns
                comp._linearize()
                # add 5 to number of runs to cover the 5 uncolored output columns
                self.assertEqual(comp._nruns - start_nruns, sparsity.shape[0] + 10)
                jac = comp._jacobian._subjacs_info
                _check_partial_matrix(comp, jac, sparsity, method)


@use_tempdirs
class TestColoringImplicitFuncComp(unittest.TestCase):
    def setUp(self):
        np.random.seed(11)

    def test_partials_implicit_funccomp(self):
        for (method, direction), isplit, osplit in itertools.product([('fd', 'fwd'), ('cs', 'fwd'), ('jax', 'fwd'), ('jax', 'rev')] if jax else [('fd', 'fwd'), ('cs', 'fwd')],
        [1,2,7,19],
        [1,2,5,11]):
            with self.subTest(msg=f'{method=}_{direction=}_{isplit=}_{osplit=}'):
                prob = Problem(name=f'test_partials_implicit_funccomp_{method}_{direction}_{isplit}_{osplit}')
                model = prob.model

                if direction == 'rev':
                    mat = np.vstack((_BIGMASK, _BIGMASK)).T
                    ninputs = mat.shape[1]
                    sparsity = setup_sparsity(mat)
                else:
                    ninputs = _BIGMASK.shape[1]
                    sparsity = setup_sparsity(_BIGMASK)

                indeps, conns = setup_indeps(isplit, ninputs, 'indeps', 'comp')
                model.add_subsystem('indeps', indeps)
                if method == 'jax':
                    sparsity = jnp.array(sparsity)
                comp = model.add_subsystem('comp', SparseFuncCompImplicit(sparsity, method,
                                                                        isplit=isplit, osplit=osplit))
                comp.declare_coloring('*', method=method)

                for conn in conns:
                    model.connect(*conn)

                prob.setup(check=False, mode=direction)
                prob.set_solver_print(level=0)
                prob.run_model()

                comp.run_linearize()
                prob.run_model()
                comp.run_linearize()
                jac = comp._jacobian._subjacs_info
                _check_partial_matrix(comp, jac, sparsity, method)


@use_tempdirs
class TestColoringSemitotals(unittest.TestCase):

    def setUp(self):
        np.random.seed(11)

    def test_simple_semitotals(self):
        self.skipTest('Semi-total coloring currently not supported.')
        for method, isplit, osplit, sparse_partials in itertools.product(['fd', 'cs'],[1,2,19],[1,2,11],[True, False]):
            with self.subTest(f'{method=} {isplit=} {osplit=} {sparse_partials=}'):
                probname = f'test_simple_semitotals_{method}_{isplit}_{osplit}_{sparse_partials}'
                prob = Problem(name=probname)
                model = prob.model

                sparsity = setup_sparsity(_BIGMASK)
                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'sub.comp')

                model.add_subsystem('indeps', indeps)
                sub = model.add_subsystem('sub', CounterGroup())
                sub.declare_coloring('*', method=method)
                sub.add_subsystem('comp', SparseCompExplicit(sparsity, method, isplit=isplit, osplit=osplit,
                                                             sparse_partials=sparse_partials))

                for conn in conns:
                    model.connect(*conn)

                for i in range(isplit):
                    model.add_design_var('indeps.x%d' % i)

                for i in range(osplit):
                    model.sub.comp.add_constraint('y%d' % i)

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                prob.driver._compute_totals()  # this is when the dynamic coloring update happens

                start_nruns = sub._nruns
                prob.driver._compute_totals()
                _check_partial_matrix(sub, sub._jacobian._subjacs_info, sparsity, method)
                self.assertEqual(sub._nruns - start_nruns, 10)

    def test_simple_semitotals_static(self):
        self.skipTest('Semi-total coloring currently not supported.')

        for method, isplit, osplit in itertools.product(['fd', 'cs'],
                                                        [1,2,19],
                                                        [1,2,11]):
            with self.subTest(msg=f'{method=} {isplit=} {osplit=}'):
                probname = f'test_simple_partials_implicit_static_{method}_{isplit}_{osplit}'

                prob = Problem(name=probname)
                model = prob.model

                sparsity = setup_sparsity(_BIGMASK)
                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'sub.comp')

                model.add_subsystem('indeps', indeps)
                sub = model.add_subsystem('sub', Group())
                comp = sub.add_subsystem('comp', SparseCompExplicit(sparsity, method, isplit=isplit, osplit=osplit))

                for conn in conns:
                    model.connect(*conn)

                for i in range(isplit):
                    model.add_design_var('indeps.x%d' % i)

                for i in range(osplit):
                    model.sub.comp.add_constraint('y%d' % i)

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()
                coloring = sub._compute_coloring(wrt_patterns='comp.x*', method=method)[0]
                sub._save_coloring(coloring)

                # now create a second problem and use the static coloring
                prob = Problem(name=probname)
                model = prob.model

                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'sub.comp')

                model.add_subsystem('indeps', indeps)
                sub = model.add_subsystem('sub', Group())
                comp = sub.add_subsystem('comp', SparseCompExplicit(sparsity, method, isplit=isplit, osplit=osplit))

                for conn in conns:
                    model.connect(*conn)

                for i in range(isplit):
                    model.add_design_var('indeps.x%d' % i)

                for i in range(osplit):
                    model.sub.comp.add_constraint('y%d' % i)

                sub.declare_coloring(wrt='comp.x*', method=method)
                sub.use_fixed_coloring()

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                start_nruns = comp._nruns
                prob.driver._compute_totals()

                nruns = comp._nruns - start_nruns
                self.assertEqual(nruns, 10)
                _check_partial_matrix(sub, sub._jacobian._subjacs_info, sparsity, method)

    def test_semitotals_unsupported(self):
        prob = Problem(name='test_semitotals_unsupported')
        model = prob.model

        sparsity = setup_sparsity(_BIGMASK)
        indeps, conns = setup_indeps(1, _BIGMASK.shape[1], 'indeps', 'sub.comp')

        model.add_subsystem('indeps', indeps)
        sub = model.add_subsystem('sub', CounterGroup())
        sub.declare_coloring('*', method='fd')
        sub.add_subsystem('comp', SparseCompExplicit(sparsity, 'fd', isplit=1, osplit=1,
                                                     sparse_partials=False))
        for conn in conns:
            model.connect(*conn)

        with self.assertRaises(RuntimeError) as err:
            prob.setup(check=False)

        expected_msg = "'sub' <class CounterGroup>: semi-total coloring is currently not supported."
        self.assertEqual(str(err.exception), expected_msg)


@use_tempdirs
class TestColoring(unittest.TestCase):

    def setUp(self):
        np.random.seed(11)

    def test_partials_explicit_shape_bug(self):
        for method in ['fd', 'cs']:
            with self.subTest(msg=f'{method=}'):
                prob = Problem(name=f'test_partials_explicit_shape_bug_{method}')
                model = prob.model

                # create sparsity with last row and col all zeros.
                # bug caused an exception when we created a COO matrix without supplying shape
                mask = np.array(
                        [[1, 0, 0, 1, 1, 1, 0],
                        [0, 1, 0, 1, 0, 1, 0],
                        [0, 1, 0, 1, 1, 1, 0],
                        [1, 0, 0, 0, 0, 1, 0],
                        [0, 0, 0, 0, 0, 0, 0]]
                    )

                isplit = 2
                sparsity = setup_sparsity(mask)
                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                                    isplit=isplit, osplit=2))
                comp.declare_coloring('x*', method=method)

                for conn in conns:
                    model.connect(*conn)

                prob.setup(check=False, mode='fwd')

                prob.set_val('indeps.x0', [1.03, 1.04, 1.05, 1.06])
                prob.set_val('indeps.x1', [1.06, 1.07, 1.08])

                prob.set_solver_print(level=0)
                prob.run_model()

                comp._linearize()
                jac = comp._jacobian._subjacs_info
                _check_partial_matrix(comp, jac, sparsity, method)

    def test_partials_min_improvement(self):
        prob = Problem(name='test_partials_min_improvement')
        model = prob.model

        mask = np.array(
                [[1, 0, 1, 0, 1, 1],
                 [1, 1, 1, 0, 0, 1],
                 [0, 1, 0, 1, 1, 0],
                 [1, 0, 1, 0, 0, 1]]
            )

        isplit = 2
        sparsity = setup_sparsity(mask)
        indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

        model.add_subsystem('indeps', indeps)
        comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, 'cs',
                                                              isplit=isplit, osplit=2))
        comp.declare_coloring('x*', method='cs', min_improve_pct=20)

        for conn in conns:
            model.connect(*conn)

        prob.setup(check=False, mode='fwd')

        prob.set_solver_print(level=0)
        prob.run_model()
        with assert_warning(UserWarning, "'comp' <class SparseCompExplicit>: Coloring was deactivated.  Improvement of 16.7% was less than min allowed (20.0%)."):
            prob.model._linearize(None)

        start_nruns = comp._nruns
        comp._linearize()
        # verify we're doing a solve for each column
        self.assertEqual(6, comp._nruns - start_nruns)

        self.assertEqual(comp._coloring_info.coloring, None)
        self.assertEqual(comp._coloring_info['static'], None)

        jac = comp._jacobian._subjacs_info
        _check_partial_matrix(comp, jac, sparsity, 'cs')

    def test_partials_min_improvement_reuse(self):
        prob = Problem(name='test_partials_min_improvement_reuse')
        model = prob.model

        mask = np.array(
                [[1, 0, 1, 0, 1, 1],
                 [1, 1, 1, 0, 0, 1],
                 [0, 1, 0, 1, 1, 0],
                 [1, 0, 1, 0, 0, 1]]
            )

        isplit = 2
        sparsity = setup_sparsity(mask)
        indeps, _ = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

        model.add_subsystem('indeps', indeps)

        comps = []
        for i in range(3):
            cname = 'comp%d' % i
            comp = model.add_subsystem(cname, SparseCompExplicit(sparsity, 'cs',
                                                                isplit=isplit, osplit=2))
            comp.declare_coloring('x*', method='cs', min_improve_pct=20, per_instance=False)
            comps.append(comp)
            _, conns = setup_indeps(isplit, mask.shape[1], 'indeps', cname)

            for conn in conns:
                model.connect(*conn)

        prob.setup(check=False, mode='fwd')
        prob.set_solver_print(level=0)
        prob.run_model()

        for i, comp in enumerate(comps):
            if i == 0:
                with assert_warning(UserWarning, "'comp0' <class SparseCompExplicit>: Coloring was deactivated.  Improvement of 16.7% was less than min allowed (20.0%)."):
                    comp._linearize()

            start_nruns = comp._nruns
            comp._linearize()
            self.assertEqual(6, comp._nruns - start_nruns)
            self.assertEqual(comp._coloring_info.coloring, None)
            self.assertEqual(comp._coloring_info['static'], None)

            jac = comp._jacobian._subjacs_info
            _check_partial_matrix(comp, jac, sparsity, 'cs')

    @unittest.skipUnless(OPTIMIZER, 'requires pyoptsparse SLSQP.')
    def test_simple_totals(self):
        for method in ['fd', 'cs']:
            with self.subTest(msg=f'{method=}'):
                prob = Problem(name=f'test_simple_totals_{method}')
                model = prob.model = CounterGroup()
                prob.driver = pyOptSparseDriver(optimizer='SLSQP')
                prob.driver.declare_coloring()

                mask = np.array(
                    [[1, 0, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [1, 0, 0, 0, 0],
                    [0, 1, 1, 0, 0]]
                )

                isplit = 2
                sparsity = setup_sparsity(mask)
                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                model.add_subsystem('comp', SparseCompExplicit(sparsity, method, isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'comp.x0')
                model.connect('indeps.x1', 'comp.x1')
                model.declare_coloring('*', method=method, step=1e-6 if method=='fd' else None)

                model.comp.add_objective('y0', index=0)  # pyoptsparse SLSQP requires a scalar objective, so pick index 0
                model.comp.add_constraint('y1', lower=[1., 2.])
                model.add_design_var('indeps.x0', lower=np.ones(3), upper=np.ones(3)+.1)
                model.add_design_var('indeps.x1', lower=np.ones(2), upper=np.ones(2)+.1)
                model.approx_totals(method=method)
                prob.setup(check=False, mode='fwd')

                prob.set_val('indeps.x0', [1.03, 1.04, 1.05])
                prob.set_val('indeps.x1', [1.06, 1.07])

                prob.set_solver_print(level=0)
                prob.run_driver()  # need this to trigger the dynamic coloring

                prob.driver._total_jac = None

                start_nruns = model._nruns
                derivs = prob.compute_totals()
                _check_total_matrix(model, derivs, sparsity[[0,3,4],:], method)
                nruns = model._nruns - start_nruns
                self.assertEqual(nruns, 3)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_simple_totals_min_improvement(self):
        for optim in [pyOptSparseDriver, ScipyOptimizeDriver]:
            with self.subTest(msg=f'{optim=}'):
                str_optim = 'pyOptSparseDriver' if optim is \
                    pyOptSparseDriver else 'ScipyOptimizeDrier'
                prob = Problem(name=f'test_simple_totals_min_improvement_{str_optim}')
                model = prob.model = CounterGroup()
                prob.driver = optim(optimizer='SLSQP')

                prob.driver.declare_coloring()

                mask = np.array(
                    [[1, 0, 1, 1, 1],
                    [1, 1, 0, 1, 1],
                    [0, 1, 1, 1, 1],
                    [1, 0, 1, 0, 0],
                    [0, 1, 1, 0, 1]]
                )

                isplit = 2
                sparsity = setup_sparsity(mask)
                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                model.add_subsystem('comp', SparseCompExplicit(sparsity, 'cs', isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'comp.x0')
                model.connect('indeps.x1', 'comp.x1')

                model.comp.add_objective('y0', index=0)  # pyoptsparse SLSQP requires a scalar objective, so pick index 0
                model.comp.add_constraint('y1', lower=[1., 2.])
                model.add_design_var('indeps.x0', lower=np.ones(3), upper=np.ones(3)+.1)
                model.add_design_var('indeps.x1', lower=np.ones(2), upper=np.ones(2)+.1)
                model.approx_totals(method='cs')
                model.declare_coloring(min_improve_pct=25., method='cs')
                prob.setup(check=False, mode='fwd')

                prob.set_val('indeps.x0', [1.03, 1.04, 1.05])
                prob.set_val('indeps.x1', [1.06, 1.07])

                prob.set_solver_print(level=0)

                with assert_warning(UserWarning, "<model> <class CounterGroup>: Coloring was deactivated.  Improvement of 20.0% was less than min allowed (25.0%)."):
                    prob.run_driver()  # need this to trigger the dynamic coloring

                prob.driver._total_jac = None

                start_nruns = model._nruns
                derivs = prob.compute_totals()
                nruns = model._nruns - start_nruns
                _check_total_matrix(model, derivs, sparsity[[0,3,4],:], 'cs')
                self.assertEqual(nruns, 5)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_totals_over_implicit_comp(self):
        for method in ['fd', 'cs']:
            with self.subTest(msg=f'{method=}'):
                prob = Problem(name=f'test_totals_over_implicit_comp_{method}')
                model = prob.model = CounterGroup()
                prob.driver = pyOptSparseDriver(optimizer='SLSQP')
                prob.driver.declare_coloring()

                mask = np.array(
                    [[1, 0, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [1, 0, 0, 0, 0],
                    [0, 1, 1, 0, 0]]
                )

                isplit = 2
                sparsity = setup_sparsity(mask)
                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.nonlinear_solver = NonlinearBlockGS()
                model.add_subsystem('indeps', indeps)
                model.add_subsystem('comp', SparseCompImplicit(sparsity, method, isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'comp.x0')
                model.connect('indeps.x1', 'comp.x1')

                model.comp.add_objective('y0', index=1)
                model.comp.add_constraint('y1', lower=[1., 2.])
                model.add_design_var('indeps.x0', lower=np.ones(3), upper=np.ones(3)+.1)
                model.add_design_var('indeps.x1', lower=np.ones(2), upper=np.ones(2)+.1)

                model.approx_totals(method=method)

                prob.setup(check=False, mode='fwd')

                prob.set_val('indeps.x0', [1.03, 1.04, 1.05])
                prob.set_val('indeps.x1', [1.06, 1.07])

                prob.set_solver_print(level=0)
                prob.run_driver()  # need this to trigger the dynamic coloring

                prob.driver._total_jac = None

                start_nruns = model._nruns
                derivs = prob.driver._compute_totals()
                self.assertEqual(model._nruns - start_nruns, 3)
                rows = [1,3,4]
                _check_total_matrix(model, derivs, sparsity[rows, :], method)

    @unittest.skipUnless(OPTIMIZER, 'requires pyoptsparse SLSQP.')
    def test_totals_of_wrt_indices(self, method='cs', sparse_partials=True):
        for method, sparse_partials in itertools.product(['fd', 'cs'],[True, False]):
            with self.subTest(msg=f'{method=} {sparse_partials=}'):
                prob = Problem(name=f'test_totals_of_wrt_indices_{method}_{sparse_partials}')
                model = prob.model = CounterGroup()
                prob.driver = pyOptSparseDriver(optimizer='SLSQP')
                prob.driver.declare_coloring()

                mask = np.array(
                    [[1, 0, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [1, 0, 0, 0, 0],
                    [0, 1, 1, 0, 0]]
                )

                isplit=2
                sparsity = setup_sparsity(mask)
                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                               isplit=isplit, osplit=2,
                                                               sparse_partials=sparse_partials))

                model.connect('indeps.x0', 'comp.x0')
                model.connect('indeps.x1', 'comp.x1')

                model.comp.add_objective('y0', index=1)
                model.comp.add_constraint('y1', lower=[1., 2.])
                model.add_design_var('indeps.x0',  indices=[0,2], lower=np.ones(2), upper=np.ones(2)+.1)
                model.add_design_var('indeps.x1', lower=np.ones(2), upper=np.ones(2)+.1)

                model.approx_totals(method=method)

                prob.setup(check=False, mode='fwd')

                prob.set_val('indeps.x0', [1.03, 1.04, 1.05])
                prob.set_val('indeps.x1', [1.06, 1.07])

                prob.set_solver_print(level=0)
                prob.run_driver()  # need this to trigger the dynamic coloring

                prob.driver._total_jac = None

                start_nruns = model._nruns
                derivs = prob.driver._compute_totals()  # colored

                self.assertEqual(model._nruns - start_nruns, 2)
                cols = [0,2,3,4]
                rows = [1,3,4]
                _check_total_matrix(model, derivs, sparsity[rows, :][:, cols], method)

    def test_no_solver_linearize(self):
        # this raised a singularity error before the fix
        class Get_val_imp(ImplicitComponent):
            def initialize(self):
                self.options.declare('size',default=1)
            def setup(self):
                size = self.options['size']
                self.add_output('state',val=5.0*np.ones(size))
                self.add_input('bar',val=1.4*np.ones(size))
                self.add_input('foobar')

                self.nonlinear_solver = NewtonSolver()
                self.linear_solver = DirectSolver()

                self.nonlinear_solver.options['iprint'] = 2
                self.nonlinear_solver.options['maxiter']=1
                self.nonlinear_solver.options['solve_subsystems']=False

                self.declare_partials(of="*", wrt="*", method="fd")

                self.declare_coloring(method="fd", num_full_jacs=2)

            def apply_nonlinear(self,inputs,outputs,residuals):
                foo = outputs['state']
                bar = inputs['bar']

                area_ratio = 1 / foo * np.sqrt(1/(bar+1)* (1/foo**2))

                residuals['state']=inputs['foobar']-area_ratio

        size = 3

        ivc = IndepVarComp()
        ivc.add_output('bar',val=1.4*np.ones(size))
        ivc.add_output('foobar',val=40)


        p = Problem()
        p.model.add_subsystem('ivc',ivc,promotes=['*'])
        p.model.add_subsystem('comp_check', Get_val_imp(size=size))
        p.model.connect('bar','comp_check.bar')
        p.setup()
        p.run_model()


@use_tempdirs
class TestStaticColoring(unittest.TestCase):

    def setUp(self):
        np.random.seed(11)

    def test_partials_explicit_shape_bug(self):
        for method in ['fd', 'cs']:
            with self.subTest(msg=f'{method=}'):
                prob = Problem(name=f'test_partials_explicit_shape_bug_static_{method}')
                model = prob.model

                # create sparsity with last row and col all zeros.
                # bug happened when we created a COO matrix without supplying shape
                mask = np.array(
                        [[1, 0, 0, 1, 1, 1, 0],
                        [0, 1, 0, 1, 0, 1, 0],
                        [0, 1, 0, 1, 1, 1, 0],
                        [1, 0, 0, 0, 0, 1, 0],
                        [0, 0, 0, 0, 0, 0, 0]]
                    )

                isplit = 2
                sparsity = setup_sparsity(mask)
                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                                    isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'comp.x0')
                model.connect('indeps.x1', 'comp.x1')

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()
                coloring = comp._compute_coloring(wrt_patterns='x*', method=method)[0]
                comp._save_coloring(coloring)

                # now make a second problem to use the coloring
                _clear_problem_names()
                prob = Problem(name=f'test_partials_explicit_shape_bug_static_{method}')
                model = prob.model
                indeps = IndepVarComp()
                indeps.add_output('x0', np.ones(4))
                indeps.add_output('x1', np.ones(3))

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                                    isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'comp.x0')
                model.connect('indeps.x1', 'comp.x1')

                comp.declare_coloring(wrt='x*', method=method)
                comp.use_fixed_coloring()
                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                comp._linearize()
                jac = comp._jacobian._subjacs_info
                _check_partial_matrix(comp, jac, sparsity, method)

    def test_simple_totals_static(self):
        for method, isplit, osplit in itertools.product(['fd', 'cs'],
                                                        [1,2,19],
                                                        [1,2,11]):
            with self.subTest(msg=f'{method=}_{isplit=}_{osplit=}'):
                prob = Problem(name=f'test_simple_totals_static_{method}_{isplit}_{osplit}')
                model = prob.model

                sparsity = setup_sparsity(_BIGMASK)
                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method, isplit=isplit, osplit=osplit))

                for conn in conns:
                    model.connect(*conn)

                for i in range(isplit):
                    model.add_design_var('indeps.x%d' % i)

                for i in range(osplit):
                    model.comp.add_constraint('y%d' % i)

                model.approx_totals(method=method)
                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()
                # try just manually computing the coloring here instead of using declare_coloring
                coloring = compute_total_coloring(prob)
                model._save_coloring(coloring)

                # new Problem, loading the coloring we just computed
                _clear_problem_names()
                prob = Problem(name=f'test_simple_totals_static_{method}_{isplit}_{osplit}')
                model = prob.model

                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method, isplit=isplit, osplit=osplit))

                for conn in conns:
                    model.connect(*conn)

                for i in range(isplit):
                    model.add_design_var('indeps.x%d' % i)

                for i in range(osplit):
                    model.comp.add_constraint('y%d' % i)

                model.approx_totals(method=method)

                model.declare_coloring(wrt='*', method=method)
                model.use_fixed_coloring()

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                start_nruns = comp._nruns
                derivs = prob.driver._compute_totals()

                nruns = comp._nruns - start_nruns
                self.assertEqual(nruns, 10)
                _check_total_matrix(model, derivs, sparsity, method)

    def test_static_totals_over_implicit_comp(self):
        for method, isplit, osplit in itertools.product(['fd', 'cs'],
                                                        [1,2,19],
                                                        [1,2,11]):
            with self.subTest(msg=f'{method=}_{isplit=}_{osplit=}'):
                prob = Problem(name=f'test_static_totals_over_implicit_comp_{method}_{isplit}_{osplit}')

                model = prob.model

                sparsity = setup_sparsity(_BIGMASK)
                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'comp')

                model.nonlinear_solver = NonlinearBlockGS()
                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompImplicit(sparsity, method, isplit=isplit, osplit=osplit))

                for conn in conns:
                    model.connect(*conn)

                for i in range(isplit):
                    model.add_design_var('indeps.x%d' % i)

                for i in range(osplit):
                    model.comp.add_constraint('y%d' % i)

                model.approx_totals(method=method)

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()
                model._save_coloring(compute_total_coloring(prob))

                _clear_problem_names()
                prob = Problem(name=f'test_static_totals_over_implicit_comp_{method}_{isplit}_{osplit}')
                model = prob.model

                indeps, conns = setup_indeps(isplit, _BIGMASK.shape[1], 'indeps', 'comp')

                model.nonlinear_solver = NonlinearBlockGS()
                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompImplicit(sparsity, method, isplit=isplit, osplit=osplit))
                for conn in conns:
                    model.connect(*conn)

                for i in range(isplit):
                    model.add_design_var('indeps.x%d' % i)

                for i in range(osplit):
                    model.comp.add_constraint('y%d' % i)

                model.approx_totals(method=method)

                model.declare_coloring(wrt='*', method=method)
                model.use_fixed_coloring()

                prob.setup(check=False, mode='fwd')

                prob.set_solver_print(level=0)
                prob.run_model()

                start_nruns = comp._nruns
                derivs = prob.driver._compute_totals()  # colored

                nruns = comp._nruns - start_nruns
                # NLBGS ends up doing a single iteration after initialization, resulting in 2
                # runs per NL solve, so we multiplly the number of colored solvers by 2
                self.assertEqual(nruns, 10 * 2)
                _check_total_matrix(model, derivs, sparsity, method)

    def test_totals_of_indices(self):
        for method in ['fd', 'cs']:
            with self.subTest(msg=f'{method=}'):
                prob = Problem(name=f'test_totals_of_indices_{method}')

                model = prob.model

                mask = np.array(
                    [[1, 0, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [1, 0, 0, 0, 0],
                    [0, 1, 1, 0, 0]]
                )

                isplit = 2
                sparsity = setup_sparsity(mask)
                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method, isplit=isplit, osplit=2))
                for conn in conns:
                    model.connect(*conn)

                model.comp.add_constraint('y0', indices=[0,2])
                model.comp.add_constraint('y1')
                model.add_design_var('indeps.x0')
                model.add_design_var('indeps.x1')
                model.approx_totals(method=method)

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()
                model._save_coloring(compute_total_coloring(prob))

                _clear_problem_names()
                prob = Problem(name=f'test_totals_of_indices_{method}')
                model = prob.model

                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method, isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'comp.x0')
                model.connect('indeps.x1', 'comp.x1')

                model.comp.add_constraint('y0', indices=[0,2])
                model.comp.add_constraint('y1')
                model.add_design_var('indeps.x0')
                model.add_design_var('indeps.x1')
                model.approx_totals(method=method)

                model.declare_coloring(wrt='*', method=method)
                model.use_fixed_coloring()

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                start_nruns = comp._nruns
                derivs = prob.driver._compute_totals()  # colored

                nruns = comp._nruns - start_nruns
                self.assertEqual(nruns, 3)
                rows = [0,2,3,4]
                _check_total_matrix(model, derivs, sparsity[rows, :], method)

    def test_totals_wrt_indices(self):
        for method in ['fd', 'cs']:
            with self.subTest(msg=f'{method=}'):
                prob = Problem(name=f'test_totals_wrt_indices_{method}')
                model = prob.model

                mask = np.array(
                    [[1, 0, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [1, 0, 0, 0, 0],
                    [0, 1, 1, 0, 0]]
                )

                isplit = 2
                sparsity = setup_sparsity(mask)
                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                                    isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'comp.x0')
                model.connect('indeps.x1', 'comp.x1')

                model.comp.add_constraint('y0')
                model.comp.add_constraint('y1')
                model.add_design_var('indeps.x0', indices=[0,2])
                model.add_design_var('indeps.x1')
                model.approx_totals(method=method)

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()
                model._save_coloring(compute_total_coloring(prob))

                _clear_problem_names()
                prob = Problem(name=f'test_totals_wrt_indices_{method}')
                model = prob.model

                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                                        isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'comp.x0')
                model.connect('indeps.x1', 'comp.x1')

                model.comp.add_constraint('y0')
                model.comp.add_constraint('y1')
                model.add_design_var('indeps.x0', indices=[0,2])
                model.add_design_var('indeps.x1')
                model.approx_totals(method=method)

                model.declare_coloring(wrt='*', method=method)
                model.use_fixed_coloring()

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                start_nruns = comp._nruns
                derivs = prob.driver._compute_totals()  # colored

                nruns = comp._nruns - start_nruns
                # only 4 cols to solve for, but we get coloring of [[2],[3],[0,1]] so only 1 better
                self.assertEqual(nruns, 3)
                cols = [0,2,3,4]
                _check_total_matrix(model, derivs, sparsity[:, cols], method)

    def test_totals_of_wrt_indices(self):
        for method in ['fd', 'cs']:
            with self.subTest(msg=f'{method=}'):
                prob = Problem(name=f'test_totals_of_wrt_indices_{method}')
                model = prob.model

                mask = np.array(
                    [[1, 0, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [0, 1, 0, 1, 1],
                    [1, 0, 0, 0, 0],
                    [0, 1, 1, 0, 0]]
                )

                isplit = 2
                sparsity = setup_sparsity(mask)
                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                                    isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'comp.x0')
                model.connect('indeps.x1', 'comp.x1')

                model.comp.add_constraint('y0', indices=[0,2])
                model.comp.add_constraint('y1')
                model.add_design_var('indeps.x0', indices=[0,2])
                model.add_design_var('indeps.x1')

                model.approx_totals(method=method)

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()
                model._save_coloring(compute_total_coloring(prob))

                _clear_problem_names()
                prob = Problem(name=f'test_totals_of_wrt_indices_{method}')
                model = prob.model

                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                                        isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'comp.x0')
                model.connect('indeps.x1', 'comp.x1')

                model.comp.add_constraint('y0', indices=[0,2])
                model.comp.add_constraint('y1')
                model.add_design_var('indeps.x0', indices=[0,2])
                model.add_design_var('indeps.x1')
                model.approx_totals(method=method)

                model.declare_coloring(wrt='*', method=method)
                model.use_fixed_coloring()

                prob.setup(check=False, mode='fwd')

                prob.set_solver_print(level=0)
                prob.run_model()

                start_nruns = comp._nruns
                derivs = prob.driver._compute_totals()  # colored

                nruns = comp._nruns - start_nruns
                self.assertEqual(nruns, 3)
                cols = rows = [0,2,3,4]
                _check_total_matrix(model, derivs, sparsity[rows, :][:, cols], method)



@unittest.skipUnless(MPI and PETScVector, "MPI and PETSc is required.")
@use_tempdirs
class TestStaticColoringParallelCS(unittest.TestCase):
    N_PROCS = 2

    def setUp(self):
        np.random.seed(11)

    # semi-total coloring feature disabled.

    def test_simple_semitotals_all_local_vars(self):
        raise self.skipTest('Semi-total coloring currently not supported.')
        for method in ['fd', 'cs']:
            with self.subTest(msg=f'{method=}'):
                prob = Problem(name=f'test_simple_semitotals_all_local_vars_{method}')

                model = prob.model

                mask = np.array(
                    [[1, 0, 0, 1, 1],
                        [0, 1, 0, 1, 1],
                        [0, 1, 0, 1, 1],
                        [1, 0, 0, 0, 0],
                        [0, 1, 1, 0, 0]]
                )
                if prob.comm.rank == 0:
                    sparsity = setup_sparsity(mask)
                    prob.comm.bcast(sparsity, root=0)
                else:
                    sparsity = prob.comm.bcast(None, root=0)

                isplit = 2
                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                sub = model.add_subsystem('sub', Group(num_par_fd=self.N_PROCS))
                sub.approx_totals(method=method)
                comp = sub.add_subsystem('comp', SparseCompExplicit(sparsity, method, isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'sub.comp.x0')
                model.connect('indeps.x1', 'sub.comp.x1')

                model.sub.comp.add_constraint('y0')
                model.sub.comp.add_constraint('y1')
                model.add_design_var('indeps.x0')
                model.add_design_var('indeps.x1')

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()
                coloring = sub._compute_coloring(wrt_patterns='*', method=method)[0]
                sub._save_coloring(coloring)

                # make sure coloring file exists by the time we try to load the spec
                prob.comm.barrier()

                # now create a second problem and use the static coloring
                prob = Problem(name=f'test_simple_semitotals_all_local_vars_{method}')
                model = prob.model

                indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

                model.add_subsystem('indeps', indeps)
                sub = model.add_subsystem('sub', Group(num_par_fd=self.N_PROCS))
                #sub.approx_totals(method=method)
                comp = sub.add_subsystem('comp', SparseCompExplicit(sparsity, method, isplit=isplit, osplit=2))
                model.connect('indeps.x0', 'sub.comp.x0')
                model.connect('indeps.x1', 'sub.comp.x1')

                model.sub.comp.add_constraint('y0')
                model.sub.comp.add_constraint('y1')
                model.add_design_var('indeps.x0')
                model.add_design_var('indeps.x1')

                sub.declare_coloring(wrt='*', method=method)
                sub.use_fixed_coloring()

                prob.setup(check=False, mode='fwd')
                prob.set_solver_print(level=0)
                prob.run_model()

                start_nruns = comp._nruns
                #derivs = prob.driver._compute_totals()
                sub._linearize(sub._jacobian)
                nruns = comp._nruns - start_nruns
                if sub._full_comm is not None:
                    nruns = sub._full_comm.allreduce(nruns)

                _check_partial_matrix(sub, sub._jacobian._subjacs_info, sparsity, method)
                self.assertEqual(nruns, 3)

    @parameterized.expand(['fd', 'cs'], name_func=lambda func, n, p: f'{func.__name__}_{p.args[0]}')
    def test_simple_partials_implicit(self, method):
        prob = Problem(name=f'test_simple_partials_implicit_{method}')
        model = prob.model

        mask = np.array(
            [[1, 0, 0, 1, 1, 1, 0],
                [0, 1, 0, 1, 0, 1, 1],
                [0, 1, 0, 1, 1, 1, 0],
                [1, 0, 0, 0, 0, 1, 0],
                [0, 1, 1, 0, 1, 1, 1]]
        )

        om_dump('---------- bcast sparsity')
        if prob.comm.rank == 0:
            sparsity = setup_sparsity(mask)
            prob.comm.bcast(sparsity, root=0)
        else:
            sparsity = prob.comm.bcast(None, root=0)
        om_dump('---------- bcast sparsity DONE')

        isplit = 2
        indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

        model.add_subsystem('indeps', indeps)
        comp = model.add_subsystem('comp', SparseCompImplicit(sparsity, method,
                                                            isplit=isplit, osplit=2,
                                                            num_par_fd=self.N_PROCS))
        model.connect('indeps.x0', 'comp.x0')
        model.connect('indeps.x1', 'comp.x1')

        om_dump('---------------- problem setup')

        prob.setup(check=False, mode='fwd')
        prob.set_solver_print(level=0)

        om_dump('--------------- run model')

        prob.run_model()
        om_dump('--------------- run model DONE')

        coloring = comp._compute_coloring(wrt_patterns='x*', method=method)[0]
        om_dump('--------------- compute_coloring DONE')

        comp._save_coloring(coloring)
        om_dump('--------------- save_coloring DONE')

        # make sure coloring file exists by the time we try to load the spec
        prob.comm.barrier()
        om_dump('--------------- barrier DONE')

        _clear_problem_names()
        prob = Problem(name=f'test_simple_partials_implicit_{method}')
        model = prob.model

        indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

        model.add_subsystem('indeps', indeps)
        comp = model.add_subsystem('comp', SparseCompImplicit(sparsity, method,
                                                                isplit=isplit, osplit=2,
                                                                num_par_fd=self.N_PROCS))
        model.connect('indeps.x0', 'comp.x0')
        model.connect('indeps.x1', 'comp.x1')

        comp.declare_coloring(wrt='x*', method=method)
        comp.use_fixed_coloring()

        om_dump('--------------- 2nd problem setup')
        prob.setup(check=False, mode='fwd')
        prob.set_solver_print(level=0)
        om_dump('---------------  2nd problem run_model')
        prob.run_model()
        om_dump('---------------  2nd problem run_model DONE')

        start_nruns = comp._nruns
        om_dump('---------------  comp linearize')
        comp._linearize()   # colored
        om_dump('---------------  comp linearize DONE')
        # number of runs = ncolors + number of outputs (only input columns were colored here)
        nruns = comp._nruns - start_nruns
        if comp._full_comm is not None:
            nruns = comp._full_comm.allreduce(nruns)
        om_dump('---------------  allreduce DONE')
        self.assertEqual(nruns, 5 + sparsity.shape[0])
        om_dump('---------------  assert check DONE')

        jac = comp._jacobian._subjacs_info
        _check_partial_matrix(comp, jac, sparsity, method)
        om_dump('---------------  check_partial_matrix DONE')

    @parameterized.expand(['fd', 'cs'], name_func=lambda func, n, p: f'{func.__name__}_{p.args[0]}')
    def test_simple_partials_explicit(self, method):
        from openmdao.utils.reports_system import clear_report_registry
        prob = Problem(name=f'test_simple_partials_explicit_{method}')
        model = prob.model

        mask = np.array(
                [[1, 0, 0, 1, 1, 1, 0],
                    [0, 1, 0, 1, 0, 1, 1],
                    [0, 1, 0, 1, 1, 1, 0],
                    [1, 0, 0, 0, 0, 1, 0],
                    [0, 1, 1, 0, 1, 1, 1]]
            )

        if prob.comm.rank == 0:
            sparsity = setup_sparsity(mask)
            prob.comm.bcast(sparsity, root=0)
        else:
            sparsity = prob.comm.bcast(None, root=0)
        om_dump('---------------  sparsity bcast DONE')

        isplit = 2
        indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

        model.add_subsystem('indeps', indeps)
        comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                            isplit=isplit, osplit=2,
                                                            num_par_fd=self.N_PROCS))
        model.connect('indeps.x0', 'comp.x0')
        model.connect('indeps.x1', 'comp.x1')

        prob.setup(check=False, mode='fwd')
        om_dump('---------------  setup DONE')
        prob.set_solver_print(level=0)
        prob.run_model()
        om_dump('---------------  run_model DONE')
        coloring = comp._compute_coloring(wrt_patterns='x*', method=method)[0]
        om_dump('---------------  compute_coloring DONE')
        comp._save_coloring(coloring)
        om_dump('---------------  save_coloring DONE')

        # make sure coloring file exists by the time we try to load the spec
        prob.comm.barrier()
        om_dump('---------------  barrier DONE')

        # now create a new problem and use the previously generated coloring
        _clear_problem_names()
        clear_report_registry()
        prob = Problem(name=f'test_simple_partials_explicit_{method}')
        model = prob.model

        indeps, conns = setup_indeps(isplit, mask.shape[1], 'indeps', 'comp')

        model.add_subsystem('indeps', indeps)
        comp = model.add_subsystem('comp', SparseCompExplicit(sparsity, method,
                                                                isplit=isplit, osplit=2,
                                                                num_par_fd=self.N_PROCS))
        model.connect('indeps.x0', 'comp.x0')
        model.connect('indeps.x1', 'comp.x1')

        comp.declare_coloring(wrt='x*', method=method)
        comp.use_fixed_coloring()
        prob.setup(check=False, mode='fwd')
        om_dump('---------------  2nd prob setup DONE')
        prob.set_solver_print(level=0)
        prob.run_model()
        om_dump('---------------  2nd prob run_model DONE')

        start_nruns = comp._nruns
        comp._linearize()
        om_dump('---------------  comp linearize DONE')
        nruns = comp._nruns - start_nruns
        if comp._full_comm is not None:
            nruns = comp._full_comm.allreduce(nruns)
        om_dump('---------------  allreduce DONE')
        self.assertEqual(nruns, 5)
        om_dump('---------------  assert check DONE')

        jac = comp._jacobian._subjacs_info
        _check_partial_matrix(comp, jac, sparsity, method)
        om_dump('---------------  check_partial_matrix DONE')


if __name__ == '__main__':
    unittest.main()



"""
Contains test groups for cycles with easily verified values/derivatives.

The Group contains the following components:
    - 'first'
    - (num_comp - 2 copies of) 'middle_i'
    - 'last'

Components other than the 'last' component take an input vector and rotate that vector by a
specified angle in the plane spanned by [1, 0, ..., 0, 1] and [0, 1, ..., 1, 0]. 'first' takes a
parameter 'psi' that corresponds to the initial rotation. The middle components use the rotation
angle specified by 'theta'. All components (including 'first' and 'last') have an output 'theta_out'
that is used to pass the 'theta' angle to the next component. The goal of this group is to determine
a value for 'theta' such that
    psi + (num_comp-1)*theta = 0 (mod 2*pi),
i.e., an angle 'theta' such that rotations by the non-first components would return the vector to
its original location. The 'last' component computes a new value for 'theta_out' that has 1/2 the
error as the input 'theta'. 'last.theta_out' is connected to 'first.theta' to form a cycle and the
reduction of error by 1/2 ensures Gauss-Seidel type methods will converge.

To provide support for multiple variables/variable sizes, the vector is constructed from the inputs
by taking the inputs in order (e.g. x_0, x_1, ...), flattening the array, and concatenating those
arrays into one vector. The outputs y_i are constructed in the reverse manner.

Note: 'theta' is unique only up to equivalence mod (2*pi)/(num_comp - 1). Test authors should not
depend on particular values of 'theta' (or 'x_i'/'y_i' values) without taking this in to account.
"""

import numpy as np

import openmdao.api as om
from openmdao.test_suite.components.cycle_comps import PSI, \
    ExplicitCycleComp, ExplicitFirstComp, ExplicitLastComp
from openmdao.test_suite.groups.parametric_group import ParametericTestGroup


class CycleGroup(ParametericTestGroup):
    """
    Group with a cycle. Derivatives and values are known.
    """

    def initialize(self):
        self.default_params.update({
            'connection_type': ['implicit', 'explicit'],
            'partial_type': ['array', 'sparse', 'aij'],
            'partial_method': ['exact', 'fd', 'cs'],
            'num_comp': [3, 2],
            'num_var': [3, 1],
            'var_shape': [(2, 3), (3,)],
        })

        self.options.declare('num_comp', types=int, default=2,
                             desc='Total number of components')
        self.options.declare('num_var', types=int, default=1,
                             desc='Number of variables per component')
        self.options.declare('var_shape', default=(3,),
                             desc='Shape of each variable')
        self.options.declare('connection_type', default='explicit',
                             values=['explicit', 'implicit'],
                             desc='How to connect variables.')
        self.options.declare('partial_type', default='array',
                             values=['array', 'sparse', 'aij'],
                             desc='type of partial derivatives')
        self.options.declare('partial_method', default='exact',
                             values=('exact', 'fd', 'cs'),
                             desc='Method used to solve derivatives (exact, fd, cs).')

    def setup(self):
        num_comp = self.options['num_comp']
        if num_comp < 2:
            raise ValueError('Number of components must be at least 2.')

        self.num_var = num_var = self.options['num_var']
        self.var_shape = var_shape = self.options['var_shape']

        self.size = num_var * np.prod(var_shape)
        if self.size < 3:
            raise ValueError('Product of num_var and var_shape must be at least 3.')

        connection_type = self.options['connection_type']

        first_class = ExplicitFirstComp
        middle_class = ExplicitCycleComp
        last_class = ExplicitLastComp

        self._generate_components(connection_type, first_class, middle_class, last_class, num_comp)

        theta_name = 'last.theta_out' if connection_type == 'explicit' else \
            'theta_{}'.format(num_comp)

        self.total_of = ['last.x_norm2', theta_name]
        self.total_wrt = ['psi_comp.psi']
        self.expected_totals = {
            ('last.x_norm2', 'psi_comp.psi'): [[0.]],
            (theta_name, 'psi_comp.psi'): [[-1. / (num_comp - 1)]],
        }

        expected_theta = (2 * np.pi - PSI) / (num_comp - 1)
        self.expected_values = {
            theta_name : expected_theta,
            'last.x_norm2': 0.5 * self.size,
        }

    def _generate_components(self, conn_type, first_class, middle_class, last_class, num_comp):
        first_name = 'first'
        last_name = 'last'
        var_shape = self.options['var_shape']
        num_var = self.options['num_var']
        comp_args = {
            'var_shape': var_shape,
            'num_var': num_var,
            'jacobian_type': self.options['jacobian_type'],
            'partial_type': self.options['partial_type'],
            'connection_type': conn_type,
            'partial_method': self.options['partial_method'],
            'num_comp': self.options['num_comp']
        }

        self.add_subsystem('psi_comp', om.IndepVarComp('psi', PSI))
        indep_var_comp = self.add_subsystem('x0_comp', om.IndepVarComp())
        for i in range(num_var):
            indep_var_comp.add_output('x_{0}'.format(i), val=np.ones(var_shape))

        idx = 0

        first_comp = first_class(index=idx, **comp_args)
        first_comp._init_parameterized()
        self.add_subsystem(first_name, first_comp,
                           promotes_inputs=first_comp._cycle_promotes_in,
                           promotes_outputs=first_comp._cycle_promotes_out)
        prev_name = first_name

        connection_variables = [('y_{0}'.format(i), 'x_{0}'.format(i)) for i in range(num_var)]
        connection_variables.append(('theta_out', 'theta'))

        # Middle Subsystems
        for idx in range(1, num_comp - 1):
            current_name = 'middle_{0}'.format(idx)
            middle_comp = middle_class(index=idx, **comp_args)
            middle_comp._init_parameterized()
            self.add_subsystem(current_name, middle_comp,
                               promotes_inputs=middle_comp._cycle_promotes_in,
                               promotes_outputs=middle_comp._cycle_promotes_out)

            if conn_type == 'explicit':
                self._explicit_connections(prev_name, current_name, connection_variables)

            prev_name = current_name

        # Final Subsystem
        last_comp = last_class(index=idx+1, **comp_args)
        last_comp._init_parameterized()
        self.add_subsystem(last_name, last_comp,
                           promotes_inputs=last_comp._cycle_promotes_in,
                           promotes_outputs=last_comp._cycle_promotes_out)

        if conn_type == 'explicit':
            self._explicit_connections(prev_name, last_name, connection_variables)
            self._explicit_connections(last_name, first_name, [('theta_out', 'theta')])
        else:
            theta_out = last_comp._cycle_names['theta_out']
            theta_in = first_comp._cycle_names['theta']
            self.connect(theta_out, theta_in)

        self.connect('psi_comp.psi', first_name + '.psi')
        self.connect('psi_comp.psi', last_name + '.psi')

        if conn_type == 'explicit':
            var_name = first_name + '.x_{0}'
        else:
            var_name = 'x_0_{0}'

        for i in range(num_var):
            self.connect('x0_comp.x_{0}'.format(i), var_name.format(i))

    def _explicit_connections(self, prev_name, current_name, vars):
        for out_var, in_var in vars:
            self.connect(
                '{0}.{1}'.format(prev_name, out_var),
                '{0}.{1}'.format(current_name, in_var)
            )



from openmdao.api import OptionsDictionary

import unittest

from openmdao.utils.assert_utils import assert_warning, assert_no_warning
from openmdao.utils.om_warnings import OMDeprecationWarning

from openmdao.core.explicitcomponent import ExplicitComponent


def check_even(name, value):
    if value % 2 != 0:
        raise ValueError("Option '%s' with value %s is not an even number." % (name, value))


class TestOptionsDict(unittest.TestCase):

    def setUp(self):
        self.dict = OptionsDictionary()

    def test_reprs(self):
        class MyComp(ExplicitComponent):
            pass

        my_comp = MyComp()

        self.dict.declare('test', values=['a', 'b'], desc='Test integer value')
        self.dict.declare('flag', default=False, types=bool)
        self.dict.declare('comp', default=my_comp, types=ExplicitComponent)
        self.dict.declare('long_desc', types=str,
                          desc='This description is long and verbose, so it '
                               'takes up multiple lines in the options table.')

        self.assertEqual(repr(self.dict), repr(self.dict._dict))

        self.assertEqual(self.dict.__str__(width=89).strip(), """
=========  ============  =================  =====================  ======================
Option     Default       Acceptable Values  Acceptable Types       Description           
=========  ============  =================  =====================  ======================
comp       MyComp        N/A                ['ExplicitComponent']                        
flag       False         [True, False]      ['bool']                                     
long_desc  **Required**  N/A                ['str']                This description is   
                                                                   long and verbose, so  
                                                                   it takes up multiple  
                                                                   lines in the options  
                                                                   table.                
test       **Required**  ['a', 'b']         N/A                    Test integer value    
=========  ============  =================  =====================  ======================
""".strip())

        # if the table can't be represented in specified width, then we get the full width version
        self.assertEqual(self.dict.__str__(width=40).strip(), """
=========  ============  =================  =====================  =========================================================================================
Option     Default       Acceptable Values  Acceptable Types       Description                                                                              
=========  ============  =================  =====================  =========================================================================================
comp       MyComp        N/A                ['ExplicitComponent']                                                                                           
flag       False         [True, False]      ['bool']                                                                                                        
long_desc  **Required**  N/A                ['str']                This description is long and verbose, so it takes up multiple lines in the options table.
test       **Required**  ['a', 'b']         N/A                    Test integer value                                                                       
=========  ============  =================  =====================  =========================================================================================
""".strip())

    def test_to_table(self):
        class MyComp(ExplicitComponent):
            pass

        my_comp = MyComp()

        self.dict.declare('test', values=['a', 'b'], desc='Test integer value')
        self.dict.declare('flag', default=False, types=bool)
        self.dict.declare('comp', default=my_comp, types=ExplicitComponent)
        self.dict.declare('long_desc', types=str,
                          desc='This description is long and verbose, so it '
                               'takes up multiple lines in the options table.')

        expected = \
"""
| Option    | Default      | Acceptable Values | Acceptable Types      | Description                                                                               |
| :-------- | :----------- | :---------------- | :-------------------- | :---------------------------------------------------------------------------------------- |
| comp      | MyComp       | N/A               | ['ExplicitComponent'] |                                                                                           |
| flag      | False        | [True, False]     | ['bool']              |                                                                                           |
| long_desc | **Required** | N/A               | ['str']               | This description is long and verbose, so it takes up multiple lines in the options table. |
| test      | **Required** | ['a', 'b']        | N/A                   | Test integer value                                                                        |
"""
        self.assertEqual(self.dict.to_table(fmt='github').strip(), expected.strip())

    def test_deprecation_col(self):
        class MyComp(ExplicitComponent):
            pass

        my_comp = MyComp()

        self.dict.declare('test', values=['a', 'b'], desc='Test integer value')
        self.dict.declare('flag', default=False, types=bool)
        self.dict.declare('comp', default=my_comp, types=ExplicitComponent)
        self.dict.declare('long_desc', types=str,
                          desc='This description is long and verbose, so it '
                               'takes up multiple lines in the options table.',
                          deprecation='This option is deprecated')

        expected = \
"""
| Option    | Default      | Acceptable Values | Acceptable Types      | Description                                                                               | Deprecation               |
| :-------- | :----------- | :---------------- | :-------------------- | :---------------------------------------------------------------------------------------- | :------------------------ |
| comp      | MyComp       | N/A               | ['ExplicitComponent'] |                                                                                           | N/A                       |
| flag      | False        | [True, False]     | ['bool']              |                                                                                           | N/A                       |
| long_desc | **Required** | N/A               | ['str']               | This description is long and verbose, so it takes up multiple lines in the options table. | This option is deprecated |
| test      | **Required** | ['a', 'b']        | N/A                   | Test integer value                                                                        | N/A                       |
"""

        self.assertEqual(self.dict.to_table(fmt='github').strip(), expected.strip())

        my_comp = MyComp()

        self.dict.declare('test', values=['a', 'b'], desc='Test integer value')
        self.dict.declare('flag', default=False, types=bool)
        self.dict.declare('comp', default=my_comp, types=ExplicitComponent)
        self.dict.declare('long_desc', types=str,
                          desc='This description is long and verbose, so it '
                               'takes up multiple lines in the options table.')

        expected = \
"""
| Option    | Default      | Acceptable Values | Acceptable Types      | Description                                                                               |
| :-------- | :----------- | :---------------- | :-------------------- | :---------------------------------------------------------------------------------------- |
| comp      | MyComp       | N/A               | ['ExplicitComponent'] |                                                                                           |
| flag      | False        | [True, False]     | ['bool']              |                                                                                           |
| long_desc | **Required** | N/A               | ['str']               | This description is long and verbose, so it takes up multiple lines in the options table. |
| test      | **Required** | ['a', 'b']        | N/A                   | Test integer value                                                                        |
"""

        self.assertEqual(self.dict.to_table(fmt='github').strip(), expected.strip())

    def test_type_checking(self):
        self.dict.declare('test', types=int, desc='Test integer value')

        self.dict['test'] = 1
        self.assertEqual(self.dict['test'], 1)

        with self.assertRaises(TypeError) as context:
            self.dict['test'] = ''

        expected_msg = "Value ('') of option 'test' has type 'str', " \
                       "but type 'int' was expected."
        self.assertEqual(expected_msg, str(context.exception))

        # multiple types are allowed
        self.dict.declare('test_multi', types=(int, float), desc='Test multiple types')

        self.dict['test_multi'] = 1
        self.assertEqual(self.dict['test_multi'], 1)
        self.assertEqual(type(self.dict['test_multi']), int)

        self.dict['test_multi'] = 1.0
        self.assertEqual(self.dict['test_multi'], 1.0)
        self.assertEqual(type(self.dict['test_multi']), float)

        with self.assertRaises(TypeError) as context:
            self.dict['test_multi'] = ''

        expected_msg = "Value ('') of option 'test_multi' has type 'str', " \
                       "but one of types ('int', 'float') was expected."
        self.assertEqual(expected_msg, str(context.exception))

        # make sure bools work and allowed values are populated
        self.dict.declare('flag', default=False, types=bool)
        self.assertEqual(self.dict['flag'], False)
        self.dict['flag'] = True
        self.assertEqual(self.dict['flag'], True)

        meta = self.dict._dict['flag']
        self.assertEqual(meta['values'], (True, False))

    def test_allow_none(self):
        self.dict.declare('test', types=int, allow_none=True, desc='Test integer value')
        self.dict['test'] = None
        self.assertEqual(self.dict['test'], None)

    def test_type_and_values(self):
        # Test with only type_
        self.dict.declare('test1', types=int)
        self.dict['test1'] = 1
        self.assertEqual(self.dict['test1'], 1)

        # Test with only values
        self.dict.declare('test2', values=['a', 'b'])
        self.dict['test2'] = 'a'
        self.assertEqual(self.dict['test2'], 'a')

        # Test with both type_ and values
        with self.assertRaises(Exception) as context:
            self.dict.declare('test3', types=int, values=['a', 'b'])
        self.assertEqual(str(context.exception),
                         "'types' and 'values' were both specified for option 'test3'.")

    def test_check_valid_template(self):
        # test the template 'check_valid' function
        from openmdao.utils.options_dictionary import check_valid
        self.dict.declare('test', check_valid=check_valid)

        with self.assertRaises(ValueError) as context:
            self.dict['test'] = 1

        expected_msg = "Option 'test' with value 1 is not valid."
        self.assertEqual(expected_msg, str(context.exception))

    def test_isvalid(self):
        self.dict.declare('even_test', types=int, check_valid=check_even)
        self.dict['even_test'] = 2
        self.dict['even_test'] = 4

        with self.assertRaises(ValueError) as context:
            self.dict['even_test'] = 3

        expected_msg = "Option 'even_test' with value 3 is not an even number."
        self.assertEqual(expected_msg, str(context.exception))

    def test_unnamed_args(self):
        with self.assertRaises(KeyError) as context:
            self.dict['test'] = 1

        # KeyError ends up with an extra set of quotes.
        expected_msg = "\"Option 'test' cannot be set because it has not been declared.\""
        self.assertEqual(expected_msg, str(context.exception))

    def test_contains(self):
        self.dict.declare('test')

        contains = 'undeclared' in self.dict
        self.assertTrue(not contains)

        contains = 'test' in self.dict
        self.assertTrue(contains)

    def test_update(self):
        self.dict.declare('test', default='Test value', types=object)

        obj = object()
        self.dict.update({'test': obj})
        self.assertIs(self.dict['test'], obj)

    def test_update_extra(self):
        with self.assertRaises(KeyError) as context:
            self.dict.update({'test': 2})

        # KeyError ends up with an extra set of quotes.
        expected_msg = "\"Option 'test' cannot be set because it has not been declared.\""
        self.assertEqual(expected_msg, str(context.exception))

    def test_get_missing(self):
        with self.assertRaises(KeyError) as context:
            self.dict['missing']

        expected_msg = "Option 'missing' has not been declared."
        self.assertEqual(expected_msg, context.exception.args[0])

    def test_get_default(self):
        obj_def = object()
        obj_new = object()

        self.dict.declare('test', default=obj_def, types=object)

        self.assertIs(self.dict['test'], obj_def)

        self.dict['test'] = obj_new
        self.assertIs(self.dict['test'], obj_new)

    def test_values(self):
        obj1 = object()
        obj2 = object()
        self.dict.declare('test', values=[obj1, obj2])

        self.dict['test'] = obj1
        self.assertIs(self.dict['test'], obj1)

        with self.assertRaises(ValueError) as context:
            self.dict['test'] = object()

        expected_msg = ("Value \(<object object at 0x[0-9A-Fa-f]+>\) of option 'test' is not one of \[<object object at 0x[0-9A-Fa-f]+>,"
                        " <object object at 0x[0-9A-Fa-f]+>\].")
        self.assertRegex(str(context.exception), expected_msg)

    def test_read_only(self):
        opt = OptionsDictionary(read_only=True)
        opt.declare('permanent', 3.0)

        with self.assertRaises(KeyError) as context:
            opt['permanent'] = 4.0

        expected_msg = ("Tried to set read-only option 'permanent'.")
        self.assertRegex(str(context.exception), expected_msg)

    def test_bounds(self):
        self.dict.declare('x', default=1.0, lower=0.0, upper=2.0)

        with self.assertRaises(ValueError) as context:
            self.dict['x'] = 3.0

        expected_msg = "Value (3.0) of option 'x' exceeds maximum allowed value of 2.0."
        self.assertEqual(str(context.exception), expected_msg)

        with self.assertRaises(ValueError) as context:
            self.dict['x'] = -3.0

        expected_msg = "Value (-3.0) of option 'x' is less than minimum allowed value of 0.0."
        self.assertEqual(str(context.exception), expected_msg)

    def test_undeclare(self):
        # create an entry in the dict
        self.dict.declare('test', types=int)
        self.dict['test'] = 1

        # prove it's in the dict
        self.assertEqual(self.dict['test'], 1)

        # remove entry from the dict
        self.dict.undeclare("test")

        # prove it is no longer in the dict
        with self.assertRaises(KeyError) as context:
            self.dict['test']

        expected_msg = "Option 'test' has not been declared."
        self.assertEqual(expected_msg, context.exception.args[0])

    def test_deprecated_option(self):
        msg = 'Option "test1" is deprecated.'
        self.dict.declare('test1', deprecation=msg)

        # test double set
        with assert_warning(OMDeprecationWarning, msg):
            self.dict['test1'] = None
        # Should only generate warning first time
        with assert_no_warning(OMDeprecationWarning, msg):
            self.dict['test1'] = None

        # Also test set and then get
        msg = 'Option "test2" is deprecated.'
        self.dict.declare('test2', deprecation=msg)

        with assert_warning(OMDeprecationWarning, msg):
            self.dict['test2'] = None
        # Should only generate warning first time
        with assert_no_warning(OMDeprecationWarning, msg):
            self.dict['test2']

    def test_deprecated_tuple_option(self):
        msg = 'Option "test1" is deprecated. Use "foo" instead.'
        self.dict.declare('test1', deprecation=(msg, 'foo'))
        self.dict.declare('foo')

        # test double set
        with assert_warning(OMDeprecationWarning, msg):
            self.dict['test1'] = 'xyz'
        # Should only generate warning first time
        with assert_no_warning(OMDeprecationWarning, msg):
            self.dict['test1'] = 'zzz'

        with assert_no_warning(OMDeprecationWarning, msg):
            option = self.dict['test1']
        with assert_no_warning(OMDeprecationWarning):
            option2 = self.dict['foo']
        self.assertEqual(option, option2)

        # Also test set and then get
        msg = 'Option "test2" is deprecated. Use "foo2" instead.'
        self.dict.declare('test2', deprecation=(msg, 'foo2'))
        self.dict.declare('foo2')

        with assert_warning(OMDeprecationWarning, msg):
            self.dict['test2'] = 'abcd'
        # Should only generate warning first time
        with assert_no_warning(OMDeprecationWarning, msg):
            option = self.dict['test2']
        with assert_no_warning(OMDeprecationWarning):
            option2 = self.dict['foo2']
        self.assertEqual(option, option2)

        # test bad alias
        msg = 'Option "test3" is deprecated. Use "foo3" instead.'
        self.dict.declare('test3', deprecation=(msg, 'foo3'))

        with self.assertRaises(KeyError) as context:
            self.dict['test3'] = 'abcd'

        expected_msg = "Can't find aliased option 'foo3' for deprecated option 'test3'."
        self.assertEqual(context.exception.args[0], expected_msg)

    def test_supported_option_name(self):
        opt = OptionsDictionary()

        # We will continue to support ":" in option names.
        with assert_no_warning(OMDeprecationWarning):
            opt.declare('foo:bar', 1.0)

    def test_context_manager(self):
        options = OptionsDictionary()
        options.declare('foo', values=['a', 'b', 'c'], default=None, allow_none=True)
        options.declare('bar', types=(float, int))

        options['foo'] = 'b'
        options['bar'] = 3.14

        self.assertEqual(options['foo'], 'b')
        self.assertAlmostEqual(options['bar'], 3.14)

        with options.temporary(foo='c', bar=5):
            self.assertEqual(options['foo'], 'c')
            self.assertEqual(options['bar'], 5)
            with options.temporary(foo='a'):
                self.assertEqual(options['foo'], 'a')
            self.assertEqual(options['foo'], 'c')
            self.assertEqual(options['bar'], 5)

        self.assertEqual(options['foo'], 'b')
        self.assertAlmostEqual(options['bar'], 3.14)

    def test_call(self):
        options = OptionsDictionary()
        options.declare('foo', values=['a', 'b', 'c'], default=None, allow_none=True)
        options.declare('bar', types=(float, int))

        options['foo'] = 'b'
        options['bar'] = 3.14

        self.assertEqual(options['foo'], 'b')
        self.assertAlmostEqual(options['bar'], 3.14)
        options.set(foo='c', bar=5)
        self.assertEqual(options['foo'], 'c')
        self.assertEqual(options['bar'], 5)


if __name__ == "__main__":
    unittest.main()



"""
Optionally import numba and provide a dummy jit decorator if it is not available.

If numba is not available, this module provides a dummy jit decorator that simply returns the
original function.
"""

try:
    import numba         # noqa: F401
    from numba import *  # noqa: F403
except ImportError:
    numba = None
    prange = range

    # If numba is not available, just write a dummy jit wrapper.
    # Code will still run at a significant performance hit.
    def jit(f, *args, **kwargs):
        """
        Return original function.

        Parameters
        ----------
        f : function
            The function to be decorated.
        """
        return f



import unittest


class TestWarnings(unittest.TestCase):

    def setUp(self):
        """
        Ensure that OpenMDAO warnings are using their default filter action.
        """
        import openmdao.api as om
        om.reset_warnings()

    def test_doc_with_units_warning(self):
        """
        Test nominal UnitsWarning.
        """
        import openmdao.api as om
        import warnings

        class AComp(om.ExplicitComponent):

            def initialize(self):
                pass

            def setup(self):
                self.add_input('a', shape=(10,), units='m')
                self.add_input('x', shape=(10,), units='1/s')
                self.add_input('b', shape=(10,), units='m/s')

                self.add_output('y', shape=(10,), units='m')
                self.add_output('z', shape=(10,), units='m/s')

                self.declare_coloring(wrt='*', form='cs')

            def compute(self, inputs, outputs):
                outputs['y'] = inputs['a'] * inputs['x'] + inputs['b']
                outputs['z'] = inputs['b'] * inputs['x']

        p = om.Problem()

        p.model.add_subsystem('a_comp', AComp())
        p.model.add_subsystem('exec_comp',
                              om.ExecComp('foo = y + z',
                                          y={'shape': (10,)},
                                          z={'shape': (10,)},
                                          foo={'shape': (10,)}))

        p.model.connect('a_comp.y', 'exec_comp.y')
        p.model.connect('a_comp.z', 'exec_comp.z')
        p.driver.declare_coloring()

        p.setup()

        with warnings.catch_warnings(record=True) as w:
            p.setup()
            unit_warnings = [wm for wm in w if wm.category is om.UnitsWarning]
            assert (len(unit_warnings) == 2)

    def test_doc_ignore_units_warning(self):
        """
        Test the ability to ignore UnitsWarning
        """
        import warnings
        import openmdao.api as om

        class AComp(om.ExplicitComponent):

            def initialize(self):
                pass

            def setup(self):
                self.add_input('a', shape=(10,), units='m')
                self.add_input('x', shape=(10,), units='1/s')
                self.add_input('b', shape=(10,), units='m/s')

                self.add_output('y', shape=(10,), units='m')
                self.add_output('z', shape=(10,), units='m/s')

                self.declare_coloring(wrt='*', form='cs')

            def compute(self, inputs, outputs):
                outputs['y'] = inputs['a'] * inputs['x'] + inputs['b']
                outputs['z'] = inputs['b'] * inputs['x']

        p = om.Problem()

        p.model.add_subsystem('a_comp', AComp())
        p.model.add_subsystem('exec_comp',
                              om.ExecComp('foo = y + z',
                                          y={'shape': (10,)},
                                          z={'shape': (10,)},
                                          foo={'shape': (10,)}))

        p.model.connect('a_comp.y', 'exec_comp.y')
        p.model.connect('a_comp.z', 'exec_comp.z')
        p.driver.declare_coloring()

        with warnings.catch_warnings(record=True) as w:
            warnings.filterwarnings('ignore', category=om.UnitsWarning)

            p.setup()
            unit_warnings = [wm for wm in w if wm.category is om.UnitsWarning]
            assert (len(unit_warnings) == 0)

    def test_doc_error_on_openmdao_warning(self):
        """
        Test the ability to raise a UnitWarning to an error.
        """
        import warnings
        import openmdao.api as om

        class AComp(om.ExplicitComponent):

            def initialize(self):
                pass

            def setup(self):
                self.add_input('a', shape=(10,), units='m')
                self.add_input('x', shape=(10,), units='1/s')
                self.add_input('b', shape=(10,), units='m/s')

                self.add_output('y', shape=(10,), units='m')
                self.add_output('z', shape=(10,), units='m/s')

                self.declare_coloring(wrt='*', form='cs')

            def compute(self, inputs, outputs):
                outputs['y'] = inputs['a'] * inputs['x'] + inputs['b']
                outputs['z'] = inputs['b'] * inputs['x']

        p = om.Problem(name='error_on_openmdao_warning')

        p.model.add_subsystem('a_comp', AComp())
        p.model.add_subsystem('exec_comp',
                              om.ExecComp('foo = y + z',
                                          y={'shape': (10,)},
                                          z={'shape': (10,)},
                                          foo={'shape': (10,)}))

        p.model.connect('a_comp.y', 'exec_comp.y')
        p.model.connect('a_comp.z', 'exec_comp.z')
        p.driver.declare_coloring()

        with warnings.catch_warnings():
            warnings.filterwarnings('error', category=om.OpenMDAOWarning)

            with self.assertRaises(Exception) as e:
                p.setup()

        expected = "\nCollected errors for problem 'error_on_openmdao_warning':" \
                   "\n   <model> <class Group>: Output 'a_comp.y' with units of 'm' is connected to " \
                   "input 'exec_comp.y' which has no units."

        self.assertEqual(expected, str(e.exception), )



"""
Various functions to make it easier to build test models.
"""

import time
import numpy as np

from openmdao.core.group import Group
from openmdao.core.parallel_group import ParallelGroup
from openmdao.core.explicitcomponent import ExplicitComponent
from openmdao.core.indepvarcomp import IndepVarComp
from openmdao.utils.array_utils import evenly_distrib_idxs


class FloatFactory(object):
    def __init__(self, shape=()):
        self.shape = shape
        self.size = np.prod(shape) if shape else 1

    def __call__(self, iotype, idx):
        if self.shape:
            if iotype == 'input':
                return np.ones(self.shape)
            else:
                return np.zeros(self.shape)
        else:
            if iotype == 'input':
                return float(1.0)
            else:
                return float(0.0)


class DynComp(ExplicitComponent):
    """
    A component with a settable number of params and outputs.
    """
    def __init__(self, ninputs, noutputs,
                 nl_sleep=0.001, ln_sleep=0.001,
                 var_factory=float, vf_args=()):
        super().__init__()

        self.ninputs = ninputs
        self.noutputs = noutputs
        self.var_factory = var_factory
        self.vf_args = vf_args
        self.nl_sleep = nl_sleep
        self.ln_sleep = ln_sleep

    def setup(self):
        for i in range(self.ninputs):
            self.add_input(f'i{i}', self.var_factory(*self.vf_args))

        for i in range(self.noutputs):
            self.add_output(f'o{i}', self.var_factory(*self.vf_args))

    def compute(self, inputs, outputs):
        time.sleep(self.nl_sleep)

    def compute_partials(self, inputs, partials):
        """
        Jacobian for Sellar discipline 1.
        """
        time.sleep(self.ln_sleep)


class DynComp2(ExplicitComponent):
    """
    A component with a settable number of params and outputs.
    """
    def __init__(self, ninputs, noutputs, mult, var_factory):
        super().__init__()

        self.ninputs = ninputs
        self.noutputs = noutputs
        self.mult = mult
        self.var_factory = var_factory

    def setup(self):
        for i in range(self.ninputs):
            self.add_input(f'i{i}', self.var_factory('input', i))

        for i in range(self.noutputs):
            self.add_output(f'o{i}', self.var_factory('output', i))

        self.declare_partials('*', '*',
                              rows=np.arange(self.var_factory.size),
                              cols=np.arange(self.var_factory.size))

    def compute(self, inputs, outputs):
        outputs.set_val(inputs.asarray() * self.mult)

    def compute_partials(self, inputs, partials):
        for o in range(self.noutputs):
            for i in range(self.ninputs):
                if o == i:
                    partials[f'o{o}', f'i{i}'] = self.mult


class DynComp2Factory(object):
    def __init__(self, ninputs=10, noutputs=10, mult=1.1, var_factory=FloatFactory()):
        self.ninputs = ninputs
        self.noutputs = noutputs
        self.mult = mult
        self.var_factory = var_factory

    def __call__(self):
        return DynComp2(self.ninputs, self.noutputs, self.mult, self.var_factory)



def make_subtree(parent, nsubgroups, levels,
                 ncomps, ninputs, noutputs, nconns, var_factory=float):
    """Construct a system subtree under the given parent group."""

    if levels <= 0:
        return

    if levels == 1:  # add leaf nodes
        create_dyncomps(parent, ncomps, ninputs, noutputs, nconns,
                        var_factory=var_factory)
    else:  # add more subgroup levels
        for i in range(nsubgroups):
            g = parent.add_subsystem("G%d"%i, Group())
            make_subtree(g, nsubgroups, levels-1,
                         ncomps, ninputs, noutputs, nconns,
                         var_factory=var_factory)


def create_dyncomps(parent, ncomps, ninputs, noutputs, nconns,
                    var_factory=float):
    """Create a specified number of DynComps with a specified number
    of variables (ninputs and noutputs), and add them to the given parent
    and add the number of specified connections.
    """
    for i in range(ncomps):
        parent.add_subsystem("C%d" % i, DynComp(ninputs, noutputs, var_factory=var_factory))

        if i > 0:
            for j in range(nconns):
                parent.connect("C%d.o%d" % (i-1,j), "C%d.i%d" % (i, j))


def create_testcomps(ncomps, comp_factory):
    """
    Yield ncomps instances returned from the given component factory.
    """
    for i in range(ncomps):
        yield f"C{i}", comp_factory()


def connect_comps(model, complist, ninputs, noutputs):
    """
    Connect the components in complist.
    """
    for i in range(len(complist)):
        if i > 0:
            for j in range(min(noutputs, ninputs)):
                model.connect(f"{complist[i-1][1]}.o{j}", f"{complist[i][1]}.i{j}")


def recursive_split(parent, complist, nsplits, max_levels, level=0, path=None):
    """
    Add levels to the system tree while splitting up the current component list.

    Parameters
    ----------
    parent : Group
        The parent group.
    complist : list
        List of components to split up.
    nsplits : int
        Number of splits to make.
    max_levels : int
        Maximum number of levels in the system tree.
    level : int
        Current level in the system tree.
    path : list or None
        List of names of parents in the system tree.
    """
    if path is None:
        path = []

    sizes, offsets = evenly_distrib_idxs(nsplits, len(complist))
    subs = [complist[offset:offset + size] for size, offset in zip(sizes, offsets)]
    for subcomps in subs:
        if not subcomps:
            continue

        g = Group()
        gname = subcomps[0][0].replace('C', f'G_{level}_')
        parent.add_subsystem(gname, g)

        # if leaf node, add comps
        if level + 1 >= max_levels or len(subcomps) < nsplits:
            for i, subcomp in enumerate(subcomps):
                cname, _, comp = subcomp
                # set the pathname for connections later
                subcomp[1] = '.'.join(path + [gname, cname])
                g.add_subsystem(cname, comp)
        else:
            recursive_split(g, subcomps, nsplits, max_levels, level + 1, path + [gname])


def build_test_model(ncomps, ninputs, noutputs, splits_per_group, max_levels, shape):
    """
    Create a test problem that has design vars and responses and can compute derivatives.

    Parameters
    ----------
    ncomps : int
        Number of components in the model.
    ninputs : int
        Number of inputs to each component.
    noutputs : int
        Number of outputs from each component.
    splits_per_group : int
        Each Group has sub-Groups added based on the number of splits of the current list
        of components.
    max_levels : int
        Maximum number of levels in the system tree.
    shape : tuple
        Shape of the input and output variables.

    Returns
    -------
    Group
        The model.
    """
    model = Group()
    complist = [[name, None, comp] for name, comp in
                create_testcomps(ncomps, DynComp2Factory(ninputs, noutputs, 1.1,
                                                         FloatFactory(shape)))]
    recursive_split(model, complist, splits_per_group, max_levels)
    connect_comps(model, complist, ninputs, noutputs)
    for i in range(ninputs):
        model.add_design_var(f"{complist[0][1]}.i{i}")
    if noutputs > 0:
        model.add_objective(f"{complist[-1][1]}.o0", index=0, flat_indices=True)
    if noutputs > 1:
        for i in range(1, noutputs):
            model.add_constraint(f"{complist[-1][1]}.o{i}", lower=0.0)
    return model


if __name__ == '__main__':
    import sys
    from openmdao.core.problem import Problem
    from openmdao.devtools.debug import config_summary

    vec_size = 1000
    num_comps = 50
    pts = 2


    class SubGroup(Group):
        def setup(self):
            create_dyncomps(self, num_comps, 2, 2, 2,
                            var_factory=lambda: np.zeros(vec_size))
            cname = "C%d"%(num_comps-1)
            self.add_objective("%s.o0" % cname)
            self.add_constraint("%s.o1" % cname, lower=0.0)


    p = Problem()
    g = p.model

    if 'gmres' in sys.argv:
        from openmdao.solvers.linear.scipy_iter_solver import ScipyKrylov
        g.linear_solver = ScipyKrylov()

    g.add_subsystem("P", IndepVarComp('x', np.ones(vec_size)))

    g.add_design_var("P.x")

    par = g.add_subsystem("par", ParallelGroup())
    for pt in range(pts):
        ptname = "G%d"%pt
        ptg = par.add_subsystem(ptname, SubGroup())
        g.connect("P.x", "par.%s.C0.i0" % ptname)

    p.setup()
    p.final_setup()
    p.run_model()
    #
    from openmdao.devtools.memory import max_mem_usage
    print("mem:", max_mem_usage())

    config_summary(p)



"""
Testing the file wrapping utilities.
"""

import os

import unittest

from openmdao.utils.assert_utils import assert_near_equal, assert_equal_arrays
from openmdao.utils.testing_utils import use_tempdirs

import numpy
from numpy import array, isnan, isinf

try:
    import pyparsing
    from openmdao.utils.file_wrap import InputFileGenerator, FileParser
except ImportError:
    pyparsing = None


DIRECTORY = os.path.dirname((os.path.abspath(__file__)))


@unittest.skipUnless(pyparsing is not None, "Test requires pyparsing to be installed. (pip install pyparsing).")
@use_tempdirs
class TestCase(unittest.TestCase):
    """ Test file wrapping functions. """

    def setUp(self):
        self.templatename = 'template.dat'
        self.filename = 'filename.dat'

    def test_templated_input(self):
        template = '\n'.join([
            "Junk",
            "Anchor",
            " A 1, 2 34, Test 1e65",
            " B 4 Stuff",
            "Anchor",
            " C 77 False Inf 333.444"
        ])

        outfile = open(self.templatename, 'w')
        outfile.write(template)
        outfile.close()

        gen = InputFileGenerator()
        gen.set_template_file(self.templatename)
        gen.set_generated_file(self.filename)
        gen.set_delimiters(', ')

        gen.mark_anchor('Anchor')
        gen.transfer_var('CC', 2, 0)
        gen.transfer_var(3.0, 1, 3)
        gen.reset_anchor()
        gen.mark_anchor('Anchor', 2)
        gen.transfer_var('NaN', 1, 4)
        gen.reset_anchor()
        gen.transfer_var('55', 3, 2)
        gen.mark_anchor('C 77')
        gen.transfer_var(1.3e-37, -3, 6)
        gen.clearline(-5)
        gen.mark_anchor('Anchor', -1)
        gen.transfer_var('8.7', 1, 5)

        gen.generate()

        infile = open(self.filename, 'r')
        result = infile.read()
        infile.close()

        answer = '\n'.join([
            "",
            "Anchor",
            " A 1, 3.0 34, Test 1.3e-37",
            " B 55 Stuff",
            "Anchor",
            " C 77 False NaN 8.7"
        ])

        self.assertEqual(answer, result)

        # Test some errors
        try:
            gen.mark_anchor('C 77', 3.14)
        except ValueError as err:
            msg = "The value for occurrence must be an integer"
            self.assertEqual(str(err), msg)
        else:
            self.fail('ValueError expected')

        try:
            gen.mark_anchor('C 77', 0)
        except ValueError as err:
            msg = "0 is not valid for an anchor occurrence."
            self.assertEqual(str(err), msg)
        else:
            self.fail('ValueError expected')

        try:
            gen.mark_anchor('ZZZ')
        except RuntimeError as err:
            msg = "Could not find pattern ZZZ in template file template.dat"
            self.assertEqual(str(err), msg)
        else:
            self.fail('RuntimeError expected')

    def test_templated_input_same_anchors(self):
        template = '\n'.join([
            "CQUAD4 1 3.456",
            "CQUAD4 2 4.123",
            "CQUAD4 3 7.222",
            "CQUAD4 4"
        ])

        outfile = open(self.templatename, 'w')
        outfile.write(template)
        outfile.close()

        gen = InputFileGenerator()
        gen.set_template_file(self.templatename)
        gen.set_generated_file(self.filename)
        gen.set_delimiters(', ')

        gen.mark_anchor('CQUAD4')
        gen.transfer_var('x', 0, 2)
        gen.mark_anchor('CQUAD4')
        gen.transfer_var('y', 0, 3)
        gen.mark_anchor('CQUAD4', 2)
        gen.transfer_var('z', 0, 2)

        gen.generate()

        infile = open(self.filename, 'r')
        result = infile.read()
        infile.close()

        answer = '\n'.join([
            "CQUAD4 x 3.456",
            "CQUAD4 2 y",
            "CQUAD4 3 7.222",
            "CQUAD4 z"
        ])

        self.assertEqual(answer, result)

    def test_templated_input_arrays(self):
        template = '\n'.join([
            "Anchor",
            "0 0 0 0 0"
        ])

        outfile = open(self.templatename, 'w')
        outfile.write(template)
        outfile.close()

        gen = InputFileGenerator()
        gen.set_template_file(self.templatename)
        gen.set_generated_file(self.filename)

        gen.mark_anchor('Anchor')
        gen.transfer_array(array([1, 2, 3, 4.75, 5.0]), 1, 3, 5, sep=' ')

        gen.generate()

        infile = open(self.filename, 'r')
        result = infile.read()
        infile.close()

        answer = '\n'.join([
            "Anchor",
            "0 0 1.0 2.0 3.0 4.75 5.0"
        ])

        self.assertEqual(answer, result)

    def test_templated_input_2Darrays(self):
        template = '\n'.join([
            "Anchor",
            "0 0 0 0 0",
            "0 0 0 0 0"
        ])

        outfile = open(self.templatename, 'w')
        outfile.write(template)
        outfile.close()

        gen = InputFileGenerator()
        gen.set_template_file(self.templatename)
        gen.set_generated_file(self.filename)

        gen.mark_anchor('Anchor')
        var = array([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]])
        gen.transfer_2Darray(var, 1, 2, 1, 5)

        gen.generate()

        infile = open(self.filename, 'r')
        result = infile.read()
        infile.close()

        answer = '\n'.join([
            "Anchor",
            "1 2 3 4 5",
            "6 7 8 9 10"
        ])

        self.assertEqual(answer, result)

    def test_output_parse(self):
        data = '\n'.join([
            "Junk",
            "Anchor",
            " A 1, 2 34, Test 1e65",
            " B 4 Stuff",
            "Anchor",
            " C 77 False NaN 333.444",
            " 1,2,3,4,5",
            " Inf 1.#QNAN -1.#IND"
        ])

        outfile = open(self.filename, 'w')
        outfile.write(data)
        outfile.close()

        gen = FileParser()
        gen.set_file(self.filename)
        gen.set_delimiters(' ')

        gen.mark_anchor('Anchor')
        val = gen.transfer_var(1, 1)
        self.assertEqual(val, 'A')
        gen.reset_anchor()
        val = gen.transfer_var(3, 2)
        self.assertEqual(val, 4)
        self.assertEqual(type(val), int)
        gen.mark_anchor('Anchor',2)
        val = gen.transfer_var(1, 4)
        self.assertEqual(isnan(val), True)
        val = gen.transfer_var(3, 1)
        self.assertEqual(isinf(val), True)
        val = gen.transfer_var(3, 2)
        self.assertEqual(isnan(val), True)
        val = gen.transfer_var(3, 3)
        self.assertEqual(isnan(val), True)
        val = gen.transfer_line(-1)
        self.assertEqual(val, ' B 4 Stuff')

        # Now, let's try column delimiters
        gen.set_delimiters('columns')
        gen.mark_anchor('Anchor',-1)
        val = gen.transfer_var(1, 8, 10)
        self.assertEqual(val, 'als')
        val = gen.transfer_var(1, 17)
        self.assertEqual(val, 333.444)

        # Test some errors
        try:
            gen.mark_anchor('C 77', 3.14)
        except ValueError as err:
            msg = "The value for occurrence must be an integer"
            self.assertEqual(str(err), msg)
        else:
            self.fail('ValueError expected')

        try:
            gen.mark_anchor('C 77', 0)
        except ValueError as err:
            msg = "0 is not valid for an anchor occurrence."
            self.assertEqual(str(err), msg)
        else:
            self.fail('ValueError expected')

        try:
            gen.mark_anchor('ZZZ')
        except RuntimeError as err:
            msg = "Could not find pattern ZZZ in output file filename.dat"
            self.assertEqual(str(err), msg)
        else:
            self.fail('RuntimeError expected')

    def test_output_parse_same_anchors(self):
        data = '\n'.join([
            "CQUAD4 1 3.456",
            "CQUAD4 2 4.123",
            "CQUAD4 3 7.222",
            "CQUAD4 4"
        ])

        outfile = open(self.filename, 'w')
        outfile.write(data)
        outfile.close()

        gen = FileParser()
        gen.set_file(self.filename)
        gen.set_delimiters(' ')

        gen.mark_anchor('CQUAD4')
        val = gen.transfer_var(0, 3)
        self.assertEqual(val, 3.456)

        gen.mark_anchor('CQUAD4')
        val = gen.transfer_var(0, 3)
        self.assertEqual(val, 4.123)

        gen.mark_anchor('CQUAD4', 2)
        val = gen.transfer_var(0, 2)
        self.assertEqual(val, 4)

        gen.reset_anchor()

        gen.mark_anchor('CQUAD4', -1)
        val = gen.transfer_var(0, 2)
        self.assertEqual(val, 4)

        gen.mark_anchor('CQUAD4', -1)
        val = gen.transfer_var(0, 3)
        self.assertEqual(val, 7.222)

        gen.mark_anchor('CQUAD4', -2)
        val = gen.transfer_var(0, 3)
        self.assertEqual(val, 4.123)

    def test_output_parse_keyvar(self):
        data = '\n'.join([
            "Anchor",
            " Key1 1 2 3.7 Test 1e65",
            " Key1 3 4 3.2 ibg 0.0003",
            " Key1 5 6 6.7 Tst xxx"
        ])

        outfile = open(self.filename, 'w')
        outfile.write(data)
        outfile.close()

        gen = FileParser()
        gen.set_file(self.filename)
        gen.set_delimiters(' ')

        gen.mark_anchor('Anchor')
        val = gen.transfer_keyvar('Key1', 3)
        self.assertEqual(val, 3.7)
        val = gen.transfer_keyvar('Key1', 4, -2)
        self.assertEqual(val, 'ibg')
        val = gen.transfer_keyvar('Key1', 4, -2, -1)
        self.assertEqual(val, 'Test')

        try:
            gen.transfer_keyvar('Key1', 4, 0)
        except ValueError as err:
            msg = "The value for occurrence must be a nonzero integer"
            self.assertEqual(str(err), msg)
        else:
            self.fail('ValueError expected')

        try:
            gen.transfer_keyvar('Key1', 4, -3.4)
        except ValueError as err:
            msg = "The value for occurrence must be a nonzero integer"
            self.assertEqual(str(err), msg)
        else:
            self.fail('ValueError expected')

    def test_output_parse_array(self):
        data = '\n'.join([
            "Anchor",
            "10 20 30 40 50 60 70 80",
            "11 21 31 41 51 61 71 81",
            "Key a b c d e"
        ])

        outfile = open(self.filename, 'w')
        outfile.write(data)
        outfile.close()

        gen = FileParser()
        gen.set_file(self.filename)
        gen.set_delimiters(' ')

        gen.mark_anchor('Anchor')
        val = gen.transfer_array(1, 1, 1, 8)
        self.assertEqual(val[0], 10)
        self.assertEqual(val[7], 80)
        val = gen.transfer_array(1, 5, 2, 6)
        self.assertEqual(val[0], 50)
        self.assertEqual(val[9], 61)
        gen.mark_anchor('Key')
        val = gen.transfer_array(0, 2, 0, 6)
        self.assertEqual(val[4], 'e')
        val = gen.transfer_array(0, 2, fieldend=6)
        self.assertEqual(val[4], 'e')

        # Now, let's try column delimiters
        gen.reset_anchor()
        gen.mark_anchor('Anchor')
        gen.set_delimiters('columns')
        val = gen.transfer_array(1, 7, 1, 15)
        self.assertEqual(val[0], 30)
        self.assertEqual(val[2], 50)
        val = gen.transfer_array(1, 10, 2, 18)
        self.assertEqual(val[0], 40)
        self.assertEqual(val[5], 61)
        val = gen.transfer_array(3, 5, 3, 10)
        self.assertEqual(val[0], 'a b c')

        try:
            gen.transfer_array(1, 7, 1)
        except ValueError as err:
            msg = "fieldend is missing, currently required"
            self.assertEqual(str(err), msg)
        else:
            self.fail('ValueError expected')

    def test_output_parse_2Darray(self):
        data = '''
        Anchor
            FREQ  DELTA  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5
             Hz
             50.   1.0   30.0  34.8  36.3  36.1  34.6  32.0  28.4  23.9  18.5  12.2   5.0  -3.1 -12.3 -22.5 -34.0 -47.2 -63.7
             63.   1.0   36.5  41.3  42.8  42.6  41.1  38.5  34.9  30.4  25.0  18.7  11.5   3.4  -5.8 -16.0 -27.5 -40.7 -57.2
             80.   1.0   42.8  47.6  49.1  48.9  47.4  44.8  41.2  36.7  31.3  25.0  17.8   9.7   0.5  -9.7 -21.2 -34.4 -50.9
            100.   1.0   48.4  53.1  54.7  54.5  53.0  50.4  46.8  42.3  36.9  30.6  23.3  15.2   6.1  -4.2 -15.7 -28.9 -45.4
            125.   1.0   53.6  58.3  59.9  59.6  58.1  55.5  52.0  47.5  42.0  35.7  28.5  20.4  11.2   1.0 -10.5 -23.7 -40.2
            160.   1.0   58.9  63.7  65.2  65.0  63.5  60.9  57.3  52.8  47.4  41.0  33.8  25.7  16.5   6.3  -5.2 -18.4 -34.9
            200.   1.0   63.4  68.1  69.6  69.4  67.9  65.3  61.7  57.2  51.8  45.5  38.3  30.1  21.0  10.7  -0.8 -14.0 -30.5
            250.   1.0   67.5  72.2  73.7  73.5  72.0  69.4  65.8  61.3  55.9  49.5  42.3  34.2  25.0  14.8   3.3 -10.0 -26.5
            315.   1.0   71.3  76.1  77.6  77.4  75.8  73.2  69.7  65.1  59.7  53.4  46.1  38.0  28.8  18.6   7.1  -6.2 -22.7
            400.   1.0   74.9  79.7  81.2  81.0  79.4  76.8  73.2  68.7  63.2  56.9  49.7  41.5  32.4  22.1  10.6  -2.7 -19.2
            500.   1.0   77.9  82.7  84.2  83.9  82.4  79.8  76.2  71.6  66.2  59.8  52.6  44.4  35.3  25.0  13.5   0.2 -16.3
            630.   1.0   80.7  85.4  86.9  86.6  85.1  82.4  78.8  74.3  68.8  62.5  55.2  47.0  37.9  27.6  16.1   2.8 -13.7
            800.   1.0   83.1  87.8  89.2  89.0  87.4  84.8  81.2  76.6  71.1  64.8  57.5  49.3  40.1  29.9  18.3   5.0 -11.5
           1000.   1.0   84.9  89.6  91.1  90.8  89.2  86.6  82.9  78.4  72.9  66.5  59.2  51.0  41.8  31.5  20.0   6.6  -9.9
           1250.   1.0   86.4  91.1  92.5  92.2  90.7  88.0  84.3  79.7  74.2  67.8  60.5  52.3  43.1  32.8  21.2   7.9  -8.7
           1600.   1.0   87.6  92.3  93.7  93.4  91.8  89.1  85.4  80.8  75.2  68.8  61.5  53.3  44.0  33.7  22.1   8.7  -7.9
           2000.   1.0   88.4  93.0  94.4  94.0  92.4  89.6  85.9  81.3  75.7  69.3  61.9  53.7  44.4  34.0  22.4   9.0  -7.6
           2500.   1.0   88.7  93.3  94.6  94.2  92.6  89.8  86.1  81.4  75.8  69.3  61.9  53.6  44.3  33.9  22.2   8.8  -7.9
           3150.   1.0   88.7  93.2  94.5  94.1  92.4  89.5  85.7  81.0  75.4  68.8  61.4  53.0  43.7  33.3  21.5   8.1  -8.6
           4000.   1.0   88.3  92.7  94.0  93.5  91.7  88.8  85.0  80.2  74.5  67.9  60.4  52.0  42.5  32.0  20.2   6.7 -10.0
           5000.   1.0   87.5  91.9  93.1  92.5  90.7  87.7  83.8  78.9  73.2  66.5  58.9  50.4  40.9  30.4  18.5   4.9 -11.9
           6300.   1.0   86.5  90.8  91.9  91.2  89.3  86.2  82.2  77.3  71.4  64.6  57.0  48.4  38.8  28.1  16.2   2.5 -14.5
           8000.   1.0   85.3  89.5  90.4  89.6  87.6  84.4  80.2  75.2  69.2  62.3  54.5  45.8  36.1  25.3  13.2  -0.6 -17.7
          10000.   1.0   84.2  88.2  89.0  88.1  85.9  82.5  78.3  73.0  66.9  59.9  51.9  43.1  33.2  22.3  10.1  -3.9 -21.1
        '''

        outfile = open(self.filename, 'w')
        outfile.write(data)
        outfile.close()

        gen = FileParser()
        gen.set_file(self.filename)

        # whitespace delim; with end field
        gen.set_delimiters(' \t')
        gen.mark_anchor('Anchor')
        val = gen.transfer_2Darray(3, 2, 26, 19)
        self.assertEqual(val[0, 1], 30.0)
        self.assertEqual(val[0, 17], -63.7)
        self.assertEqual(val[1, 17], -57.2)
        self.assertEqual(val[23, 17], -21.1)
        self.assertEqual(val.shape[0], 24)
        self.assertEqual(val.shape[1], 18)

        # whitespace delim; no end field
        val = gen.transfer_2Darray(3, 2, 26)
        self.assertEqual(val[0, 1], 30.0)
        self.assertEqual(val[23, 17], -21.1)
        self.assertEqual(val.shape[0], 24)
        self.assertEqual(val.shape[1], 18)

        # column delim; with end field
        gen.set_delimiters('columns')
        val = gen.transfer_2Darray(3, 19, 26, 125)
        self.assertEqual(val[0, 1], 30.0)
        self.assertEqual(val[0, 17], -63.7)
        self.assertEqual(val[1, 17], -57.2)
        self.assertEqual(val[23, 17], -21.1)
        self.assertEqual(val.shape[0], 24)
        self.assertEqual(val.shape[1], 18)

        # column delim; no end field
        val = gen.transfer_2Darray(3, 19, 26)
        self.assertEqual(val[0, 1], 30.0)
        self.assertEqual(val[0, 17], -63.7)
        self.assertEqual(val[1, 17], -57.2)
        self.assertEqual(val[23, 17], -21.1)
        self.assertEqual(val.shape[0], 24)
        self.assertEqual(val.shape[1], 18)

        # make sure single line works
        gen.set_delimiters(' \t')
        val = gen.transfer_2Darray(5, 3, 5, 5)
        self.assertEqual(val[0, 2], 49.1)

        # Small block read
        val = gen.transfer_2Darray(7, 3, 9, 6)
        self.assertEqual(val[0, 0], 53.6)
        self.assertEqual(val[2, 0], 63.4)

        # Error messages for bad values
        try:
            gen.transfer_2Darray(7, 3, 9, 1)
        except ValueError as err:
            msg = "fieldend must be greater than fieldstart"
            self.assertEqual(str(err), msg)
        else:
            self.fail('ValueError expected')

        try:
            gen.transfer_2Darray(9, 2, 8, 4)
        except ValueError as err:
            msg = "rowend must be greater than rowstart"
            self.assertEqual(str(err), msg)
        else:
            self.fail('ValueError expected')

    def test_comment_char(self):
        # Check to see if the use of the comment
        #   characters works
        data = '\n'.join([
            "Junk",
            "CAnchor",
            " Z 11, 22 344, Test 1e65",
            " B 4 Stuff",
            "  $ Anchor",
            " Q 1, 2 34, Test 1e65",
            " B 4 Stuff",
            "Anchor",
            " A 1, 2 34, Test 1e65",
            " B 4 Stuff",
            "Anchor",
            " C 77 False NaN 333.444",
            " 1,2,3,4,5",
            " Inf 1.#QNAN -1.#IND"
        ])

        outfile = open(self.filename, 'w')
        outfile.write(data)
        outfile.close()

        # Test full line comments
        gen = FileParser(full_line_comment_char="C")
        gen.set_file(self.filename)
        gen.set_delimiters(' ')
        gen.mark_anchor('Anchor')
        val = gen.transfer_var(1, 1)
        self.assertEqual(val, 'A')

        # Test end of line comments also
        gen = FileParser(full_line_comment_char="C", end_of_line_comment_char="$")
        gen.set_file(self.filename)
        gen.set_delimiters(' ')
        gen.mark_anchor('Anchor')
        val = gen.transfer_var(1, 1)
        self.assertEqual(val, 'A')

    def test_more_delims(self):
        data = '\n'.join([
            "anchor,1.0,2.0",
            "abc=123.456",
            "c=1,2,Word,6",
            "d=C:/abc/def,a+b*c^2,(%#%),!true",
            "a^33 1.#QNAN^#$%^"
        ])

        outfile = open(self.filename, 'w')
        outfile.write(data)
        outfile.close()

        op = FileParser()
        op.set_file(self.filename)

        op.set_delimiters(' \t,=')

        op.mark_anchor('anchor')

        val = op.transfer_var(0, 1)
        self.assertEqual(val, 'anchor')
        val = op.transfer_var(0, 2)
        self.assertEqual(val, 1.0)
        val = op.transfer_var(1, 1)
        self.assertEqual(val, 'abc')
        val = op.transfer_var(1, 2)
        self.assertEqual(val, 123.456)
        val = op.transfer_var(2, 4)
        self.assertEqual(val, 'Word')
        val = op.transfer_var(2, 5)
        self.assertEqual(val, 6)
        val = op.transfer_var(3, 2)
        self.assertEqual(val, 'C:/abc/def')
        val = op.transfer_var(3, 3)
        self.assertEqual(val, 'a+b*c^2')
        val = op.transfer_var(3, 4)
        self.assertEqual(val, '(%#%)')
        val = op.transfer_var(3, 5)
        self.assertEqual(val, '!true')

        op.set_delimiters(' \t^')
        val = op.transfer_var(4, 1)
        self.assertEqual(val, 'a')
        val = op.transfer_var(4, 2)
        self.assertEqual(val, 33)
        val = op.transfer_var(4, 3)
        self.assertEqual(isnan(val), True)
        val = op.transfer_var(4, 4)
        self.assertEqual(val, '#$%')


@unittest.skipUnless(pyparsing is not None, "Test requires pyparsing to be installed. (pip install pyparsing).")
class FileGenFeature(unittest.TestCase):

    # output data for each test
    output_data = {
        "": [
            "INPUT",
            "1 2 3",
            "INPUT",
            "10.1 20.2 30.3",
            "A B C"
        ],
        "test_transfer": [
            "INPUT",
            "1 7 3",
            "INPUT",
            "10.1 20.2 30.3",
            "A B C"
        ],
        "test_transfer_2": [
            "INPUT",
            "1 7 3",
            "INPUT",
            "10.1 20.2 3.141592653589793",
            "A B C"
        ],
        "test_transfer_minus2": [
            "INPUT",
            "99999 7 3",
            "INPUT",
            "10.1 20.2 3.141592653589793",
            "A B C"
        ],
        "test_transfer_array": [
            "INPUT",
            "123 456 789",
            "INPUT",
            "10.1 20.2 3.141592653589793",
            "A B C"
        ],
        "test_transfer_stretch": [
            "INPUT",
            "11 22 33 44 55 66",
            "INPUT",
            "10.1 20.2 3.141592653589793",
            "A B C"
        ]
    }

    # the name of the preceding test in the feature doc
    prev_test = {
        "test_transfer": "",
        "test_transfer_2": "test_transfer",
        "test_transfer_minus2": "test_transfer_2",
        "test_transfer_array": "test_transfer_minus2",
        "test_transfer_stretch": "test_transfer_array"
    }

    def setUp(self):

        global parser  # global so we don't need `self.` in feature doc
        parser = InputFileGenerator()

        # the input data for each test is the output of the previous test
        prev_test = self.prev_test[self._testMethodName]
        parser._data = self.output_data[prev_test][:]

    def test_transfer(self):
        parser.mark_anchor("INPUT")
        parser.transfer_var(7, 1, 2)
        self.assertEqual(parser.generate(),
                         '\n'.join(self.output_data[self._testMethodName]))

    def test_transfer_2(self):
        parser.mark_anchor("INPUT", 2)

        my_var = 3.1415926535897932
        parser.transfer_var(my_var, 1, 3)

        self.assertEqual(parser.generate(),
                         '\n'.join(self.output_data[self._testMethodName]))

    def test_transfer_minus2(self):
        parser.reset_anchor()
        parser.mark_anchor("INPUT", -2)
        parser.transfer_var("99999", 1, 1)

        self.assertEqual(parser.generate(),
                         '\n'.join(self.output_data[self._testMethodName]))

    def test_transfer_array(self):
        from numpy import array

        array_val = array([123, 456, 789])

        parser.reset_anchor()
        parser.mark_anchor("INPUT")
        parser.transfer_array(array_val, 1, 1, 3)

        self.assertEqual(parser.generate(),
                         '\n'.join(self.output_data[self._testMethodName]))

    def test_transfer_stretch(self):
        from numpy import array

        array_val = array([11, 22, 33, 44, 55, 66])

        parser.reset_anchor()
        parser.mark_anchor("INPUT")
        parser.transfer_array(array_val, 1, 1, 3, sep=' ')

        self.assertEqual(parser.generate(),
                         '\n'.join(self.output_data[self._testMethodName]))


@unittest.skipUnless(pyparsing is not None, "Test requires pyparsing to be installed. (pip install pyparsing).")
class FileParserFeature(unittest.TestCase):

    def setUp(self):

        global parser  # global so we don't need `self.` in feature doc
        parser = FileParser()

        parser._data = [
            "LOAD CASE 1",
            "STRESS 1.3334e7 3.9342e7 NaN 2.654e5",
            "DISPLACEMENT 2.1 4.6 3.1 2.22234",
            "LOAD CASE 2",
            "STRESS 11 22 33 44 55 66",
            "DISPLACEMENT 1.0 2.0 3.0 4.0 5.0"
        ]

    def assert_equal_arrays(self, a1, a2):
        assert_equal_arrays(a1, a2)

    def test_parse_output(self):
        parser.mark_anchor("LOAD CASE")
        var = parser.transfer_var(1, 2)

        self.assertEqual((var, type(var)), (1.3334e+07, float))

    def test_parse_nan(self):
        parser.reset_anchor()
        parser.mark_anchor("LOAD CASE")
        var = parser.transfer_var(1, 4)

        from numpy import isnan
        self.assertEqual(isnan(var), True)

    def test_parse_string(self):
        parser.reset_anchor()
        parser.mark_anchor("LOAD CASE")
        var = parser.transfer_var(2, 1)

        self.assertEqual((var, type(var)), ("DISPLACEMENT", str))

    def test_parse_output_2(self):
        parser.reset_anchor()
        parser.mark_anchor("LOAD CASE", 2)
        var = parser.transfer_var(1, 2)

        self.assertEqual((var, type(var)), (11, int))

    def test_parse_output_minus2(self):
        parser.reset_anchor()
        parser.mark_anchor("LOAD CASE", -2)
        var = parser.transfer_var(1, 2)

        self.assertAlmostEqual(var, 1.3334e+07)

    def test_parse_keyvar(self):
        parser.reset_anchor()
        parser.mark_anchor("LOAD CASE 1")
        var = parser.transfer_keyvar("DISPLACEMENT", 1)

        self.assertEqual(var, 2.1)

    def test_parse_array(self):
        parser.reset_anchor()
        parser.mark_anchor("LOAD CASE")
        var = parser.transfer_array(2, 2, 2, 5)

        assert_near_equal(var, numpy.array([2.1, 4.6, 3.1, 2.22234]))

    def test_parse_array_multiline(self):
        parser.reset_anchor()
        parser.mark_anchor("LOAD CASE")
        var = parser.transfer_array(1, 3, 2, 4)

        self.assert_equal_arrays(var, numpy.array([
            '39342000.0', 'nan', '265400.0',
            'DISPLACEMENT', '2.1', '4.6', '3.1'
        ]))


@unittest.skipUnless(pyparsing is not None, "Test requires pyparsing to be installed. (pip install pyparsing).")
class FileParser2dFeature(unittest.TestCase):

    def setUp(self):

        global parser  # global so we don't need `self.` in feature doc
        parser = FileParser()

        # A way to "cheat" and do this without a file.
        parser._data = []
        parser._data.append('FREQ  DELTA  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5  -8.5')
        parser._data.append(' Hz')
        parser._data.append(' 50.   1.0   30.0  34.8  36.3  36.1  34.6  32.0  28.4  23.9  18.5  12.2')
        parser._data.append(' 63.   1.0   36.5  41.3  42.8  42.6  41.1  38.5  34.9  30.4  25.0  18.7')
        parser._data.append(' 80.   1.0   42.8  47.6  49.1  48.9  47.4  44.8  41.2  36.7  31.3  25.0')
        parser._data.append('100.   1.0   48.4  53.1  54.7  54.5  53.0  50.4  46.8  42.3  36.9  30.6')

    def assert_equal_arrays(self, a1, a2):
        assert_equal_arrays(a1, a2)

    def test_parse_array_2d(self):
        parser.reset_anchor()
        parser.mark_anchor("Hz")
        var = parser.transfer_2Darray(1, 3, 4, 12)

        self.assert_equal_arrays(var, numpy.array([
            [30.0,  34.8,  36.3,  36.1,  34.6,  32.0,  28.4,  23.9,  18.5,  12.2],
            [36.5,  41.3,  42.8,  42.6,  41.1,  38.5,  34.9,  30.4,  25.0,  18.7],
            [42.8,  47.6,  49.1,  48.9,  47.4,  44.8,  41.2,  36.7,  31.3,  25.0],
            [48.4,  53.1,  54.7,  54.5,  53.0,  50.4,  46.8,  42.3,  36.9,  30.6]
        ]))


@unittest.skipUnless(pyparsing is not None, "Test requires pyparsing to be installed. (pip install pyparsing).")
class FileParserDelimFeature(unittest.TestCase):

    def setUp(self):

        global parser  # global so we don't need `self.` in feature doc
        parser = FileParser()

        parser._data = [
            "CASE 1",
            "3,7,2,4,5,6"
        ]

    def test_parse_default_delim(self):
        parser.reset_anchor()
        parser.mark_anchor("CASE")
        var = parser.transfer_var(1, 2)

        self.assertEqual((var, type(var)), (",7,2,4,5,6", str))

    def test_parse_comma_delim(self):
        parser.reset_anchor()
        parser.mark_anchor("CASE")
        parser.set_delimiters(", ")
        var = parser.transfer_var(1, 2)

        self.assertEqual((var, type(var)), (7, int))


@unittest.skipUnless(pyparsing is not None, "Test requires pyparsing to be installed. (pip install pyparsing).")
class FileParserColumnsFeature(unittest.TestCase):

    def setUp(self):

        global parser  # global so we don't need `self.` in feature doc
        parser = FileParser()

        parser._data = [
            "CASE 1",
            "12345678901234567890",
            "TTF    3.7-9.4434967"
        ]

    def test_parse_columns(self):
        parser.reset_anchor()
        parser.mark_anchor("CASE")
        parser.set_delimiters("columns")

        var1 = parser.transfer_var(2, 3, 3)
        var2 = parser.transfer_var(2, 4, 10)
        var3 = parser.transfer_var(2, 11, 20)

        self.assertEqual((var1, var2, var3), ('F', 3.7, -9.4434967))


@unittest.skipUnless(pyparsing is not None, "Test requires pyparsing to be installed. (pip install pyparsing).")
class FileParserArrayColumnsFeature(unittest.TestCase):

    def setUp(self):

        global parser  # global so we don't need `self.` in feature doc
        parser = FileParser()

        parser._data = [
            "CASE 2",
            "123456789012345678901234567890",
            "NODE 11 22 33 COMMENT",
            "NODE 44 55 66 STUFF"
        ]

    def test_parse_columns(self):
        parser.reset_anchor()
        parser.mark_anchor("CASE 2")

        parser.set_delimiters("columns")
        var = parser.transfer_array(2, 6, 3, 13)

        parser.set_delimiters(" \t")

        assert_near_equal(var,
                         numpy.array([11., 22., 33., 44., 55., 66.]))


if __name__ == "__main__":
    unittest.main()



"""InputResidsComp provides a simple implicit component with minimal boilerplate."""

import numpy as np

from openmdao.core.implicitcomponent import ImplicitComponent


class InputResidsComp(ImplicitComponent):
    """
    Class definition for the InputResidsComp.

    Uses all inputs as residuals while allowing individual outputs that are not necessarily
    associated with a specific residual.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.

    Attributes
    ----------
    _refs : dict
        Residual ref values that are cached during calls to the overloaded add_input method.
    """

    def __init__(self, **kwargs):
        """
        Initialize the InputResidsComp.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments passed to the __init__ method of ImplicitComponent
        """
        self._refs = {}
        super().__init__(**kwargs)

    def add_input(self, name, val=1.0, shape=None, units=None, desc='', tags=None,
                  shape_by_conn=False, copy_shape=None, compute_shape=None, distributed=None,
                  ref=None):
        """
        Add an input to be used as a residual.

        Parameters
        ----------
        name : str
            Name of the variable in this component's namespace.
        val : float or list or tuple or ndarray or Iterable
            The initial value of the variable being added in user-defined units.
            Default is 1.0.
        shape : int or tuple or list or None
            Shape of this variable, only required if val is not an array. Default is None.
        units : str or None
            Units in which this input variable will be provided to the component
            during execution. Default is None, which means it is unitless.
        desc : str
            Description of the variable.
        tags : str or list of strs
            User defined tags that can be used to filter what gets listed when calling
            list_inputs and list_outputs.
        shape_by_conn : bool
            If True, shape this input to match its connected output.
        copy_shape : str or None
            If a str, that str is the name of a variable. Shape this input to match that of
            the named variable.
        compute_shape : function
            A function taking a dict arg containing names and shapes of this component's outputs
            and returning the shape of this input.
        distributed : bool
            If True, this variable is a distributed variable, so it can have different sizes/values
            across MPI processes.
        ref : float or ndarray or None
            Scaling parameter. The value in the user-defined units of this residual
            when the scaled value is 1. Default is 1.
        """
        self._refs[name] = ref
        super().add_input(name, val=val, shape=shape, units=units, desc=desc, tags=tags,
                          shape_by_conn=shape_by_conn, copy_shape=copy_shape,
                          compute_shape=compute_shape, distributed=distributed)

    def setup_residuals(self):
        """
        Delay calls for add_residual for this component.

        This method is used since input/residual sizes may not
        be known until final setup.
        """
        for name in self._var_rel_names['input']:
            meta = self._var_rel2meta[name]
            resid_name = f'resid_{name}'
            self.add_residual(resid_name, shape=meta['shape'], units=meta['units'],
                              desc=meta['desc'], ref=self._refs[name])

    def setup_partials(self):
        """
        Delay calls to declare_partials for the component.

        This method is used because input/residual sizes
        may not be known until final setup.
        """
        for name in self._var_rel_names['input']:
            resid_name = 'resid_' + name
            size = self._var_rel2meta[name]['size']
            ar = np.arange(size, dtype=int)
            self.declare_partials(of=resid_name, wrt=name, rows=ar, cols=ar, val=1.0)

    def apply_nonlinear(self, inputs, outputs, residuals):
        """
        Compute residuals given inputs and outputs.

        The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        outputs : Vector
            Unscaled, dimensional output variables read via outputs[key].
        residuals : Vector
            Unscaled, dimensional residuals written to via residuals[key].
        """
        residuals.set_val(inputs.asarray())
